/*
 *  Linux driver for Comtrol SI/IA & DeviceMaster products.
 *
 *  Written by Grant Edwards, Comtrol Corp.  1995 - 2003.
 * 
 *  Copyright (C) 1995 - 2003 by Comtrol, Inc.
 * 
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, or (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*  This is the Linux driver for the Comtrol SI/IA & DeviceMaster products.
 *  These devices are ethernet attached serial port hubs.  For details see
 *  www.comtrol.com.
 *
 *  There are 2 "devices" in this driver.
 *  1)  NSLinkctl is a char device used by the daemon nslinkd to control the hub.
 *  2)  A tty device /dev/ttySIxx used by applications to send/receive data.
 *  
 *  The daemon nslinkd is a user space program that loads the serial hub, monitors
 *  the hub, reloads it if needed and configures it.  The daemon communicates with 
 *  the hub through the NSLinkctl device created in this driver.
 *
 *  User applications send/receive serial data through the tty devices created in this
 *  driver using the device name /dev/ttySIxx.
 */

/*
 *  Kernel Synchronization:
 *
 *  This driver has multiple kernel control paths:
 *  - TTY exception handlers (calls into the driver from user mode through tty device) 
 *  - NSLinkctl exception handlers (not a factor in synchronization, except TCP functions)
 *  - Timer bottom half (tasklet), used when connected via HDLC.
 *  - TCP functions - Each hub has a corresponding user mode process, started by the daemon,
 *    which makes an IOCTL call into the NSLinkctl device.  This call causes a polling
 *    loop to run.
 *  - HDLC network callback bottom half - In HDLC mode, received packets with protocol
 *    0x11FE are sent to the driver.
 *
 * Critical data: 
 * -  nrp_table[], accessed through passed "info" pointers, is a global (static) array of 
 *    serial port state information and the xmit & rect circular buffers.  xmit/recv head, tail 
 *    and cnt are protected with a per port spinlock, as is "count".
 * -  nrp_num_ports_open, int indicating number of open ports, protected by atomic operations.
 * -  si->rtq is a sk_buff_head list holding the HDLC retransmit queue.  The list lock is 
 *    used where needed, such as when deleting multiple entries from the list.
 * -  rp_write() and rp_write_char() functions use a per port semaphore to protect against
 *    simultaneous access to the same port by more than one process.
 * -  pstate->nPutRemote and nGetRemote only written by single context.  nGetLocal protected
 *    by a spinlock.
 * -  State variable have been left unprotected, assuming that the states are self-correcting
 *    or mutually exclusive - si->state, si->hdlc_state_flags, info->remote_status.
 */

#ifndef LINUX_VERSION_CODE
#  include <linux/version.h>
#endif

#ifdef MODVERSIONS
#include <config/modversions.h>
#endif

#ifndef VERSION_CODE
#  define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,2,9) 
#  error "This kernel is too old: not supported by this file"
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
#define TTY_DRIVER_NO_DEVFS 0
#endif

#define TraceHdlc(si,ev,d) /* nooop */
#define TraceHdlcStop(si,ev,d) /* nooop */
#define TraceQueue(info,ev,d) /* nooop */

/* remember about the current version */
#if LINUX_VERSION_CODE < VERSION_CODE(2,3,0)
#  define LINUX_22
#else
#  define LINUX_24
# endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,2,18)
/* wait queue stuff changed */
#  define DECLARE_WAIT_QUEUE_HEAD(head) struct wait_queue *head = NULL
   typedef  struct wait_queue *wait_queue_head_t;
#  define init_waitqueue_head(head) (*(head)) = NULL
   typedef  struct wait_queue  wait_queue_t;
#  define init_waitqueue_entry(w,p) do {(w)->task=p; (w)->next=NULL;} while(0)
#endif 

//#if (LINUX_VERSION_CODE > VERSION_CODE(2,2,14)) && (LINUX_VERSION_CODE < VERSION_CODE(2,4,0))
//#define HavePollWait 1
//#else
//#define HavePollWait 0
//#endif

#ifndef LINUX_24
/* offer wake_up_sync as an alias for wake_up */
//#  define wake_up_sync(head) wake_up(head)
//#  define wake_up_interruptible_sync(head) wake_up_interruptible(head)
#endif

#include <linux/compiler.h>
#include <linux/module.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/proc_fs.h>
#include <linux/poll.h>
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/netdevice.h>
#include <linux/file.h>
#include <linux/inetdevice.h>
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/stat.h>
#include <net/tcp.h>
#include <linux/tcp.h>
#include <linux/net.h>
#include <linux/timer.h>
#include <linux/init.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <asm/uaccess.h>

#include <linux/slab.h>

#if !defined(__init)
#define __init /* nothing */
#endif

#if !defined(__initfunc)
#define __initfunc(x) x
#endif

#if !defined(set_current_state)
#define set_current_state(s) current->state = (s)
#endif

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
#define NET_DEVICE net_device
#else
#define NET_DEVICE device
#define dev_get_by_name(n) dev_get(n)
#endif

#include "nslink.h"
#include "nslink_int.h"
#include "version.h"

#define MAX_CMD_PKT 1408

#define tracemsg() printk(KERN_INFO "nslink: %s(%d)\n",__FUNCTION__,__LINE__)

#undef NSLINK_DEBUG_HDLC
#undef NSLINK_DEBUG_RETRANS
#undef NSLINK_DEBUG_TIMER
#undef NSLINK_DEBUG_KEEPALIVE
#undef NSLINK_DEBUG_IN_CMDS
#undef NSLINK_DEBUG_OUT_CMDS 
#undef NSLINK_DEBUG_OPEN
#undef NSLINK_DEBUG_INTR
#undef NSLINK_DEBUG_WRITE
#undef NSLINK_DEBUG_FLOW
#undef NSLINK_DEBUG_THROTTLE
#undef NSLINK_DEBUG_WAIT_UNTIL_SENT
#undef NSLINK_DEBUG_RECEIVE
#undef NSLINK_DEBUG_HANGUP
#undef NSLINK_DEBUG_NETMEM
#undef NSLINK_DEBUG_DEVICE_ID
#undef NSLINK_DEBUG_QUEUE_COUNTS

#define NSLINK_DEBUG_TRACE_HDLC 0
#define NSLINK_DEBUG_TRACE_QUEUE 0

#ifndef time_after
#define time_after(a,b)		((long)(b) - (long)(a) < 0)
#endif

#define NumElem(a) ((sizeof (a))/(sizeof (a)[0]))
#define xmit_packet(SKBTD) dev_queue_xmit(SKBTD)

/*
 *   Static Variables
 */

// Global array of config/info structs, one per hub 
static struct si_state *conc_state[MAX_NRP_SIS];

// Global array of config/info structs, one per port.  Indexed [0-n] by 
// 'xx'  from /dev/ttySIxx.
static struct nr_port *nrp_table[(MAX_NRP_SIS * MAX_NRP_PORTS)];

// Global TTY structs
static struct tty_driver nslink_driver;
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
static struct tty_struct *nslink_table[(MAX_NRP_SIS * MAX_NRP_PORTS)];
static struct termios *nslink_termios[(MAX_NRP_SIS * MAX_NRP_PORTS)];
static struct termios  *nslink_termios_locked[(MAX_NRP_SIS * MAX_NRP_PORTS)];

static int nslink_refcount = 0;
#endif

static atomic_t nrp_num_ports_open;	 /*  Number of serial ports open  */
static int control_major = 0;

static const char *ctl_name = "NSLinkctl";

static DECLARE_WAIT_QUEUE_HEAD(wake_daemon_wait);

/* Used to report to the user the RPSH-Si boxes found on the network */
static struct si_box_report si_box_rpts[MAX_NRP_SIS];
static int num_box_rpts;	/* number of valid reports */
static int rpt_loading_only;	/* if non-zero only report available boxes */

/*
 * Function declarations
 */
static void hdlc_resync(struct si_state *si);
static void hdlc_send_ack(struct si_state *si, __u8 flags);
static void hdlc_send_remote_cmds(struct si_state *si);
static inline int get_tx_cnt_remote(struct si_port_state *pstate);
static inline int get_tx_space_remote(struct si_port_state *pstate);

static void si_set_timer(struct si_state *si, unsigned long ticks);
static void si_timer(unsigned long arg);
static int init_hub_state(struct si_box_config *in, int from_user);
static int free_si_box(int boxid);
static int set_tcp(struct tcp_mode_params *in);
static int set_mac(struct mac_mode_params *in);
static int get_box_config(struct si_box_config *retinfo);
static int get_box_report(struct si_box_report *retinfo);
static void si_parse_input_pkt(struct si_state *si, unsigned char *buf, unsigned long len);

static int ctl_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);

static void nrp_wait_until_sent(struct tty_struct *tty, int timeout);
static void nrp_flush_buffer(struct tty_struct *tty);
static void nrp_start(struct tty_struct *tty);
static void configure_nr_port(struct nr_port *info);
static void send_recv_buf_to_ldisc(struct nr_port *info);
static void push_chars_to_recv_buf(struct nr_port *info,  unsigned char *cp, int len);
static void inline wake_tty_write(struct tty_struct *tty);
static void inline wake_tty(struct tty_struct *tty);
static void wake_all_tty(struct si_state *si);
static void set_otherclose_all_tty(struct si_state *si);
static void clr_otherclose_all_tty(struct si_state *si);
static void init_nr_port(struct si_state *si, int portnum, int ic_timeout, int low_latency);

static void tcp_send_remote_cmds(struct si_state *si);
static int tcp_txqueue_wait(tTcpTxQueueWait *in);
static void tcp_store_fragment(struct si_state *si, unsigned char *buf, int len);

#ifdef NSLINK_DEBUG_NETMEM
extern void show_net_buffers(void);
#endif

static char *macToStr(unsigned char *mac);
static char *siStateStr(int i);
static char *ipToStr(unsigned long ip);
static int UnitHasNo485(int modelID);

#if LINUX_VERSION_CODE > VERSION_CODE(2,4,99)
int nrp_init(void);
static void nrp_cleanup_module(void);

module_init(nrp_init);
module_exit(nrp_cleanup_module);
#endif

#ifdef MODULE_LICENSE
MODULE_LICENSE("Dual BSD/GPL");
#endif

/********************************************************************************/
/*                                                                              */
/*        HDLC FUNCTIONS                                                        */
/*                                                                              */
/********************************************************************************/

/*
 * This allocates Linux networking buffer for sending a packet.  Note
 * that the Linux layer takes care of creating the link-level
 * (ethernet) header (i.e., the first 14 bytes of the packet).  We
 * allocate 3 bytes of "network layer header" (struct comtrolhdr),
 * followed by 3 bytes of "transport layer header" which is either the
 * 3 byte header used by the Admin packet, or the 3 byte header used
 * by the HDLC layer.
 */
static struct sk_buff *prepare_packet_hdr(struct NET_DEVICE *dev, int packet_len)
{
  struct sk_buff *skb;
  int length;

  if (!dev || !(dev->flags & IFF_UP))
    return NULL;		/* Interface not up! */

  /*
   * 6 bytes for the header, plus the data plus 15 bytes so that
   * we can align the packet properly.
   */
  length = dev->hard_header_len + 6 + packet_len + 15;

  skb = alloc_skb(length + dev->hard_header_len + 15, GFP_ATOMIC);
  if (!skb)
    return 0;

  skb->dev = dev;
  skb->protocol = __constant_htons(ETH_P_COMTROL);
  skb_reserve(skb, (dev->hard_header_len + 15) & ~15);
  skb->h.raw = skb_put(skb, 6);	/* For the headers */
  return skb;
}

/*
 * Send a Product ID query packet
 *
 * If the second argument is NULL, then we're just doing a random
 * query looking for boxes.  If the second argument is non-NULL, then
 * we're querying a specific box with the goal of telling it to listen
 * to *US*.  :-)
 */
static int send_query_packet(struct NET_DEVICE *dev, struct si_state *si)
{
  struct sk_buff *skb;
  struct comtrolhdr *chdr;
  struct adminhdr *ahdr;
  struct admin_product_id_request *areq;

  skb = prepare_packet_hdr(dev, sizeof(struct admin_product_id_request));
  if (!skb)
    {
      printk(KERN_INFO "nslink: Alloc_skb failed, is %s up?\n",dev->name);
      return -ENOMEM;
    }
  chdr = (struct comtrolhdr *) skb->h.raw;
  chdr->product_id = SI_ID;
  chdr->conc_id = si ? si->conc_id : 0x00;
  chdr->packet_class = ADMIN_CLASS;
  ahdr = (struct adminhdr *) (skb->h.raw + 3);
  ahdr->len_high = 0;
  ahdr->len_low = sizeof(struct admin_product_id_request) + 1;
  ahdr->admin_type = ADMIN_PRD_ID_REQ_TYPE;
  areq = (struct admin_product_id_request *)
    skb_put(skb, sizeof(struct admin_product_id_request));
  memcpy(areq->mac_addr, dev->dev_addr, 6);
  if (si)
    areq->flags = 0;		/* You belong to me.... */
  else
    areq->flags = 1;		/* Just testing... */
  dev->hard_header(skb, dev, ETH_P_COMTROL,
		   si ? si->ether_addr : dev->broadcast,
		   dev->dev_addr, skb->len);
  xmit_packet(skb);
  return 0;
}

/*
 * hdlc_resync -- reset the hdlc state variables
 *
 * Called by hdlc_assign_ports(), this modifies si variables, but is only called
 * prior to being connected, so can be ignored when protecting critical variables.
 */
static void hdlc_resync(struct si_state *si)
{
  struct sk_buff *skb;
  unsigned long flags;

  si->in_ack_index = 0;
  si->out_snd_index = 0;
  si->next_in_index = 0;
  si->hdlc_state_flags = 0;
  atomic_set(&si->unacked_pkts, 0);

  /* Clear the retransmit queue */
  spin_lock_irqsave(&si->rtq.lock, flags);
  while ((skb = __skb_dequeue(&si->rtq)))
    {
      --si->rtqCount;
      kfree_skb(skb);
    }
  spin_unlock_irqrestore(&si->rtq.lock, flags);
}

/*
 * Prepare an skbuff with the HDLC headers
 */
static struct sk_buff *prepare_hdlc_packet(struct si_state *si, int size,
					   __u8 flags)
{
  struct sk_buff *skb;
  struct comtrolhdr *chdr;
  struct HDLChdr *hdlc_hdr;

  skb = prepare_packet_hdr(si->dev, size);
  if (!skb)
    return 0;
  chdr = (struct comtrolhdr *) skb->h.raw;
  chdr->product_id = SI_ID;
  chdr->conc_id = si->conc_id;
  chdr->packet_class = SI_CLASS;
  hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
  hdlc_hdr->hdlc_control = flags;
  if (flags & CONTROL_UFRAME)
    hdlc_hdr->out_snd_index = 0;
  else
    hdlc_hdr->out_snd_index = si->out_snd_index;
  hdlc_hdr->next_in_index = si->next_in_index;
  return skb;
}

/*
 *  This function actually sends the HDLC packet
 *  Inputs: si-> ptr to hub info, skb-> ptr to the skb buf to send
 *  HDLC only.
 */
static void send_hdlc_packet(struct si_state *si, struct sk_buff *skb)
{
  struct HDLChdr *hdlc_hdr;
  struct sk_buff *tmp;

  hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
  si->dev->hard_header(skb, si->dev, ETH_P_COMTROL, si->ether_addr,
		       si->dev->dev_addr, skb->len);

  /*
   * Set the receiver sequence number (for acknowledgement purposes)
   * Note that we've acknowledged all of our received packets.
   */
  hdlc_hdr->next_in_index = si->next_in_index;
  atomic_set(&si->unacked_pkts, 0);
  si->hdlc_state_flags &= ~HDLC_RCV_TIMER_ACTIVE;

  if (hdlc_hdr->hdlc_control & CONTROL_UFRAME)
    {
      /*
       * This is an unsequenced frame; just send it out.
       */
      xmit_packet(skb);
      return;
    }
  
  TraceHdlc(si,3,si->out_snd_index);

  /* Set the outgoing sequence number */
  hdlc_hdr->out_snd_index = si->out_snd_index++;

  /*
   * Enable the retransmission timer, and put the packet on the
   * retransmission queue.  Finally send the packet!  
   */
  if (!(si->hdlc_state_flags & HDLC_SND_TIMER_ACTIVE))
    {
      si->sender_ack_timer = jiffies + HDLC_RETRANS_TIME;
      si->hdlc_state_flags |= HDLC_SND_TIMER_ACTIVE;
      si_set_timer(si, HDLC_RETRANS_TIME);
    }

  //  Put the skb on our retransmit queue
  skb_queue_tail(&si->rtq, skb);
  ++si->rtqCount;

  //  Create a copy of the skb, transmit it over the ethernet
  tmp = skb_clone(skb, GFP_ATOMIC);
  if (tmp)
    {
      xmit_packet(tmp);
      /* Clear the receiver ack timer */
      si->hdlc_state_flags &= ~HDLC_RCV_TIMER_ACTIVE;
    }
  return;
}

/*
 * hdlc_send_ack --- send a bare ack packet
 */
static void hdlc_send_ack(struct si_state *si, __u8 flags)
{
  struct sk_buff *skb;
  unsigned char *cp;

  skb = prepare_hdlc_packet(si, 1, flags);
  if (!skb)
    return;
  cp = skb_put(skb, 1);
  *cp = 0;
  send_hdlc_packet(si, skb);
  si->hdlc_state_flags &= ~HDLC_RCV_TIMER_ACTIVE;
  TraceHdlc(si,4,si->next_in_index);
#ifdef NSLINK_DEBUG_HDLC
  printk(KERN_INFO "nslink: hdlc: [%d] tx ack %02x\n",si->conc_id,si->next_in_index);
#endif
}

/*
 * hdlc_retransmit -- retransmit the packets in the retranmission queue.
 */
static void hdlc_retransmit(struct si_state *si)
{
  struct sk_buff *skb, *tmp, *next;
  struct HDLChdr *hdlc_hdr;
  unsigned long flags;

  //  Get a pointer to the first skb in the retransmit queue.
  //  Must grab the list spinlock so no-one grabs the skb from under us.
  spin_lock_irqsave(&si->rtq.lock, flags);
  skb = skb_peek(&si->rtq);

  hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
  ++si->retransCount;
  TraceHdlc(si,6,si->retransCount);
  if (si->retransCount == 16)
    {
      printk(KERN_INFO "nslink: hdlc: [%d] 16 retries to %s\n",si->conc_id,macToStr(si->ether_addr));
      TraceHdlcStop(si,-2,0);
    }

#if defined(NSLINK_DEBUG_RETRANS) || defined(NSLINK_DEBUG_HDLC)
    {
      unsigned char h=hdlc_hdr->out_snd_index;
      unsigned char t=h;
      if ((tmp=skb_peek_tail(&si->rtq)))
        t = ((struct HDLChdr *)(tmp->h.raw+3))->out_snd_index;
      printk(KERN_INFO "nslink: hdlc: [%d] resend %02x-%02x (LastAck %02x) count=%d\n",
             si->conc_id, h, t,si->in_ack_index,si->retransCount);
    }
#endif

  //  Walk through the retransmit queue, retransmitting the packets
  while (skb && skb != (struct sk_buff *) &si->rtq)
    {
      next = skb->next;
      tmp = skb_clone(skb, GFP_ATOMIC);
      if (tmp)
        {
          hdlc_hdr = (struct HDLChdr *)(tmp->h.raw+3);
          hdlc_hdr->next_in_index = si->next_in_index;
          xmit_packet(tmp);
        }
      skb = next;
    }
  
  if (skb_peek(&si->rtq) == NULL)
    {
#ifdef NSLINK_DEBUG_HDLC
      printk(KERN_INFO "nslink: hdlc: [%d] clearing retrans flag.\n",si->conc_id);
#endif
      si->hdlc_state_flags &= ~HDLC_SND_TIMER_ACTIVE;
    }

  spin_unlock_irqrestore(&si->rtq.lock, flags);
}

/*
 *  Reset (initialize) the port status.  Only called when the port state goes to
 *  CONNECTED, therefore this function does not change critical values.
 */
static void port_resync_all(struct si_state *si)
{
  struct si_port_state *pstate;
  struct nr_port *info;
  int i;

  for (i = 0; i < si->num_ports; i++)
    {
      pstate = si->port_state + i;

      pstate->nPutRemote = 0;
      pstate->nGetRemote = 0;
      pstate->nGetLocal = 0;
      pstate->modemStatusRegister = 0;
      pstate->errorStatusRegister = 0;
      pstate->update_queue = 0;
        
      info = nrp_table[i + si->base_port];
      if (info)
	{
          TraceQueue(info,1,0);
	  info->update_baud_rate = 1;
	  info->update_control_settings = 1;
          info->update_special_charset = 1;
	  si->need_to_send = 1;
	  info->rsmode = 0x00;

	  if ((info->remote_status == REM_PORT_ASSIGNED) ||
	      (info->remote_status == REM_OPEN_PORT))
            info->remote_status = REM_OPEN_PORT;
	  else
            info->remote_status = REM_PORT_AVAIL;
	}
    }
  if (si->need_to_send)
    {
      if (si->use_tcp == 0)
        si_timer((unsigned long) si);
      else
        tcp_send_remote_cmds(si);
    }
}
/*  Called by si_parse_input_pkt() */
static void port_resync(struct nr_port *info, struct si_port_state *pstate)
{
  pstate->nPutRemote = 0;
  pstate->nGetRemote = 0;
  pstate->nGetLocal = 0;
  pstate->modemStatusRegister = 0;
  pstate->errorStatusRegister = 0;
  pstate->update_queue = 0;

  TraceQueue(info,1,1);
  
  info->update_baud_rate = 1;
  info->update_control_settings = 1;
  info->update_special_charset = 1;
  info->rsmode = 0x00;

  return;
}

/*
 *  Puts a RK_PORT_SET command in the skb buff, if necessary.  
 *  Make sure we have at least 3 bytes in the buff calling this routine!
 *  Inputs:  si-> ptr to the hub, skb-> ptr to put data in, port number to set
 *  Returns: 0 if OK, 1 on error.  
 *  HDLC mode only.
 */
static inline int hdlc_set_port(struct si_state *si, 
                                  struct sk_buff *skb, int port)
{
  unsigned char *cp;

  if (skb->pkt_type == port)
    return 0;
  if (skb_tailroom(skb) < 3)
    return 1;
  cp = skb_put(skb, 2);
  *cp++ = RK_PORT_SET;
  *cp = port;
  skb->pkt_type = port;
#ifdef NSLINK_DEBUG_OUT_CMDS
  printk(KERN_INFO "nslink: Port %d\n", port);
#endif
  return 0;
}

/*
 *  Creates an skb buf, puts the RK_ASSIGN_REQ command in it, and sends it.
 *  Inputs:  si-> ptr to the hub
 *  HDLC mode only.
 */
static void hdlc_assign_ports(struct si_state *si)
{
  struct sk_buff *skb;
  unsigned char *cp;

  skb = prepare_hdlc_packet(si, 2, CONTROL_UFRAME);
  if (!skb)
    return;
  cp = skb_put(skb, 3);
  *cp++ = RK_ASSIGN_REQ;
  *cp = 0;
  send_hdlc_packet(si, skb);
  hdlc_resync(si);
#ifdef NSLINK_DEBUG_HDLC
  printk(KERN_INFO "nslink: hdlc: [%d] sent HDLC assign\n",si->conc_id);
#endif
}


/*
 *  Puts a RK_PORT_SET command in the skb buff, then a command that has 0 bytes
 *  of associated data. Make sure we have at least 4 bytes in the buff calling this routine!
 *  Inputs:  si-> ptr to the hub, skb-> ptr to put data in, port number to set, cmd
 *  Returns: 0 if OK, 1 on error.  
 *  HDLC mode only.
 */
static inline int hdlc_0byte_cmd(struct si_state *si, 
                                   struct sk_buff *skb, int port, 
                                   __u8 cmd)
{
  unsigned char *cp;

  if (hdlc_set_port(si, skb, port))
    return 1;
  if (skb_tailroom(skb) < 2)
    return 1;
  cp = skb_put(skb, 1);
  *cp = cmd;
  return 0;
}

/*
 *  Puts a RK_PORT_SET command in the skb buff, then a command that has 1 byte
 *  of associated data. Make sure we have at least 5 bytes in the buff calling this routine!
 *  Inputs:  si-> ptr to the hub, skb-> ptr to put data in, port number to set, cmd
 *  Returns: 0 if OK, 1 on error.  
 *  HDLC mode only.
 */
static inline int hdlc_1byte_cmd(struct si_state *si, 
                                   struct sk_buff *skb, int port, 
                                   __u8 cmd, __u8 value)
{
  unsigned char *cp;

  if (hdlc_set_port(si, skb, port))
    return 1;
  if (skb_tailroom(skb) < 3)
    return 1;
  cp = skb_put(skb, 2);
  *cp++ = cmd;
  *cp = value;
  return 0;
}

/*
 *  Puts a RK_PORT_SET command in the skb buff, then a command that has 2 bytes
 *  of associated data. Make sure we have at least 6 bytes in the buff calling this routine!
 *  Inputs:  si-> ptr to the hub, skb-> ptr to put data in, port number to set, cmd
 *  Returns: 0 if OK, 1 on error.  
 *  HDLC mode only.
 */
static inline int hdlc_2byte_cmd(struct si_state *si,
                                   struct sk_buff *skb, int port, 
                                   __u8 cmd, __u16 value)
{
  unsigned char *cp;

  if (hdlc_set_port(si, skb, port))
    return 1;
  if (skb_tailroom(skb) < 4)
    return 1;
  cp = skb_put(skb, 3);
  *cp++ = cmd;
  *cp++ = value & 0xFF;
  *cp = value >> 8;
  return 0;
}

/*
 *  Puts a RK_PORT_SET command in the skb buff, then a command that has 4 bytes
 *  of associated data. Make sure we have at least 8 bytes in the buff calling this routine!
 *  Inputs:  si-> ptr to the hub, skb-> ptr to put data in, port number to set, cmd
 *  Returns: 0 if OK, 1 on error.  
 *  HDLC mode only.
 */
static inline int hdlc_4byte_cmd(struct si_state *si,
                                   struct sk_buff *skb, int port, 
                                   __u8 cmd, __u32 value)
{
  unsigned char *cp;

  if (hdlc_set_port(si, skb, port))
    return 1;
  if (skb_tailroom(skb) < 6)
    return 1;
  cp = skb_put(skb, 5);
  *cp++ = cmd;
  *cp++ = value; 
  *cp++ = value >> 8;
  *cp++ = value >> 16;
  *cp =   value >> 24;

  return 0;
}


/*
 *  Puts a RK_PORT_SET command in the skb buff, then a command that has n bytes
 *  of associated data. 
 *  Inputs:  si-> ptr to the hub, skb-> ptr to put data in, port number to set, cmd
 *  Returns: 0 if OK, 1 on error.  
 *  HDLC mode only.
 */
static inline int hdlc_Nbyte_cmd(struct si_state *si,
                                   struct sk_buff *skb, int port, 
                                   __u8 cmd, __u8 *values, int N)
{
  unsigned char *cp;

  if (hdlc_set_port(si, skb, port))
    return 1;
  if (skb_tailroom(skb) < N+2)
    return 1;
  cp = skb_put(skb, N+1);
  *cp++ = cmd;
  memcpy(cp,values,N);
  return 0;
}

/*
 *  Copies HDLC data from info->xmit_buf into a skb.
 *  Inputs: si, info, pstate -> ptrs to hub&port info, skb-> ptr to the skb buf to be filled in
 *  Returns: 0 if OK, 1 if error
 *  HDLC mode only.
 */
static inline int hdlc_xmit(struct si_state *si, struct sk_buff *skb,
                              struct nr_port *info, struct si_port_state *pstate)
{
  unsigned char *cp;
  int c, space, total_len = 0;
  struct tty_struct *tty;
  unsigned long flags;

  if (!info || (!info->xmit_cnt && !info->x_char) ||
      !get_tx_space_remote(pstate))
    return 1;

  if (hdlc_set_port(si, skb, info->portnum))
    return 1;

  if (skb_tailroom(skb) < 5)
    return 1;

  //  Put the RK_DATA_BLK command into skb
  cp = skb_put(skb, 3);
  *cp++ = RK_DATA_BLK;
  if (info->x_char)
    {
      *(skb_put(skb, 1)) = info->x_char;
      total_len++;
      pstate->nPutRemote++;
      info->x_char = 0;
      TraceQueue(info,2,1);
    }

  //  Copy the data into skb, spinlock to protect xmit_tail and xmit_cnt
  spin_lock_irqsave(&info->slock, flags);
  while (1)
    {
      c = skb_tailroom(skb) - 1;
      if (c > info->xmit_cnt)
	c = info->xmit_cnt;
      if (c > (XMIT_BUF_SIZE - info->xmit_tail))
	c = XMIT_BUF_SIZE - info->xmit_tail;
      space = get_tx_space_remote(pstate);
      if (c > space)
	c = space;
      if (c <= 0)
	break;
      memcpy(skb_put(skb, c), info->xmit_buf + info->xmit_tail, c);
      info->xmit_tail += c;
      info->xmit_tail &= XMIT_BUF_SIZE - 1;
      info->xmit_cnt -= c;
      total_len += c;
      pstate->nPutRemote += (__u16) c;
      TraceQueue(info,2,c);
    }
  spin_unlock_irqrestore(&info->slock, flags);

#ifdef NSLINK_DEBUG_OUT_CMDS
  printk(KERN_INFO "nslink: ttySI%d: sent %d bytes\n", info->line, total_len);
#endif

  //  Copy the length info into skb
  *cp++ = total_len & 0xFF;
  *cp++ = total_len >> 8;

  tty = info->tty;
  if ((info->xmit_cnt < WAKEUP_CHARS) && tty)
    wake_tty_write(tty);
  return 0;
}

/*
 *  Sends required commands to a particular hub.  Creates a skb,
 *  fills it with connect check or update timeout info based on si flags.
 *  Then goes through each port, using it's nr_port and si_port_state 
 *  object flags to do qin update, set RSmode, set MCR, IC timeout.
 */
static void hdlc_send_remote_cmds(struct si_state *si)
{
  int send_packet;  //  If set, the skb has HDLC data that needs to be sent to the hub
  int i, aborted, first_port = 0;
  struct sk_buff *skb = NULL;
  unsigned char *cp;
  struct si_port_state *pstate;
  struct nr_port *info;

retry:
  send_packet = 0;
  si->need_to_send = 0;
  aborted = 0;

  // Get an skb
  skb = prepare_hdlc_packet(si, MAX_CMD_PKT, CONTROL_IFRAME);
  if (!skb)
    return;
  skb->pkt_type = -1;
  
  // Send connect check
  if (si->need_to_send_connect_check)
    {
      char *cp = skb_put(skb,1);
      if (cp)
        *cp = RK_CONNECT_CHECK;
      si->need_to_send_connect_check = 0;
      send_packet = 1;
    }

  // Send update timeout
  if (si->update_timeout)
    {
      char *cp = skb_put(skb,5);
      *cp++ = RK_TIMEOUT_SET;
      *cp++ = si->link_timeout;
      *cp++ = si->link_timeout >> 8;
      *cp++ = si->link_timeout >> 16;
      *cp   =   si->link_timeout >> 24;
      if ( si->scan_period && (si->arch_id == 3 || si->arch_id ==2))
        {
          char *cp = skb_put(skb,2);
          *cp++ = RK_SCANPERIOD_SET;
          *cp   = si->scan_period;
        }
      si->update_timeout = 0;
      send_packet = 1;
    }

  // Go through each port in the hub, sending needed commands
  for (i = first_port; i < si->num_ports; i++)
    {

      // Get port config/info
      info = nrp_table[i + si->base_port];
      if (!info)
        continue;
      
      if (info->remote_status == REM_PORT_ASSIGNED)
	{
	  pstate = si->port_state + i;

          // Send qin update
	  if (pstate->update_queue)
	    {
	      if (hdlc_2byte_cmd(si, skb, i, RK_QIN_STATUS, pstate->nGetLocal))
		{
		  aborted = 1;
		  break;
		}
	      pstate->update_queue = 0;
	      send_packet = 1;
	    }

          /* 
	   *  Update RSmode (RS232, etc).  DeviceMaster Serial hub has only RS232 
	   */
	  if (info->rsmode != si->rsmode[i])
	    { 
 	      if (si->rsmode[i] != SI_RSMODE_232) {
		  if (UnitHasNo485(si->modelID)) {
		      printk(KERN_INFO "nslink: Error - attempting to change interface mode for port ttySI%d on a model that only has RS232 ports \n", info->line);
		      si->rsmode[i] = SI_RSMODE_232;
		  }
	      }
	      info->rsmode = si->rsmode[i];
	      if (hdlc_1byte_cmd(si, skb, i, RK_RSMODE_SET, info->rsmode))
		{
		  info->rsmode = 0;
		  aborted = 1;
		  break;
		}
	      if (info->rsmode == SI_RSMODE_485)
		{
		  /* RS485 turn on RTS toggle */
		  info->update_control_settings = 1;
		  info->control_settings &= ~SC_FLOW_RTS_MASK;
		  info->control_settings |= SC_FLOW_RTS_ARS485;
		}
	      send_packet = 1;
	    }

          // Update baud rate 
	  if (info->update_baud_rate)
	    {
#ifdef NSLINK_DEBUG_OUT_CMDS
	      printk(KERN_INFO "nslink: set baud rate to %d\n", info->baud_rate);
#endif
	      if (hdlc_4byte_cmd(si, skb, i, RK_BAUD_SET, info->baud_rate))
		{
		  aborted = 1;
		  break;
		}
	      info->update_baud_rate = 0;
	      send_packet = 1;
	    }

          //  Set charset
	  if (info->update_special_charset)
	    {
              char b[6];
#ifdef NSLINK_DEBUG_OUT_CMDS
	      printk(KERN_INFO "nslink: set specials to %02x,%02x,%02x,%02x,%02x,%02x\n",
                     info->rx_xon,info->rx_xoff,
                     info->tx_xon,info->tx_xoff,
                     0,0);
#endif
              b[0] = info->rx_xon;
              b[1] = info->rx_xoff;
              b[2] = info->tx_xon;
              b[3] = info->tx_xoff;
              b[4] = 0;
              b[5] = 0;
	      if (hdlc_Nbyte_cmd(si, skb, i, RK_SPECIAL_CHAR_SET, b, 6))
		{
		  aborted = 1;
		  break;
		}
	      info->update_special_charset = 0;
	      send_packet = 1;
	    }

          // Control set
	  if (info->update_control_settings)
	    {
#ifdef NSLINK_DEBUG_OUT_CMDSsend_hdlc_packet
	      printk(KERN_INFO "nslink: set control byte to %d\n",
		     info->control_settings);
#endif
	      if (hdlc_2byte_cmd
		  (si, skb, i, RK_CONTROL_SET, info->control_settings))
		{
		  aborted = 1;
		  break;
		}
	      info->update_control_settings = 0;
	      send_packet = 1;
	    }

          // Set mcr
	  if (info->update_mcr)
	    {
#ifdef NSLINK_DEBUG_OUT_CMDS
	      printk(KERN_INFO "nslink: set MCR to %d\n",
		     info->modem_control_register);
#endif
	      if (hdlc_2byte_cmd
		  (si, skb, i, RK_MCR_SET, info->modem_control_register))
		{
		  aborted = 1;
		  break;
		}
	      info->update_mcr = 0;
	      send_packet = 1;
	    }

          // Set IC timeout
          if (info->update_ic_timeout)
            {
#ifdef NSLINK_DEBUG_OUT_CMDS
	      printk(KERN_INFO "nslink: set IC timeousend_hdlc_packett %d\n", info->inter_char_timeout);
#endif
	      if (hdlc_4byte_cmd (si, skb, i, RK_IC_TIMEOUT_SET, info->inter_char_timeout))
		{
		  aborted = 1;
		  break;
		}
	      info->update_ic_timeout = 0;
	      send_packet = 1;
            }

          // make sure action command is last one of the setup
          // commands we send, since we use waiting for the
          // action resonse is a synchronization method

          // Action set
	  if (info->update_action)
	    {
#ifdef NSLINK_DEBUG_OUT_CMDS
	      printk(KERN_INFO "nslink: set action to %d\n", info->action_register);
#endif
	      if (hdlc_2byte_cmd
		  (si, skb, i, RK_ACTION_SET, info->action_register))
		{
		  aborted = 1;
		  break;
		}
	      info->update_action = 0;
	      info->action_register = 0;
	      send_packet = 1;
	    }

	  /*
	   * Transmit characters in our transmit buffer
	   */
	  if (info->x_char ||
	      (info->xmit_cnt &&
	       !info->tty->hw_stopped && !info->tty->stopped))
	    {
	      if (!hdlc_xmit(si, skb, info, pstate))
		send_packet = 1;
	    }
	}

      else if (info->remote_status == REM_OPEN_PORT)
        {
          if (hdlc_1byte_cmd(si, skb, i, RK_REMOTE_STATUS, REM_OPEN_PORT))
            {
              aborted = 1;
              break;
            }
          info->remote_status = REM_WAITING_OPEN;
          info->conn_refused = 0;
          send_packet = 1;
        }

      else if (info->remote_status == REM_RELEASE_PORT)
        {
          if (hdlc_1byte_cmd(si, skb, i, RK_REMOTE_STATUS, REM_RELEASE_PORT))
            {
              aborted = 1;
              break;
            }
          info->remote_status = REM_WAITING_CLOSE;
          send_packet = 1;
	}
      
      if (info->send_force_release)
        {
          if (hdlc_0byte_cmd(si,skb,i,RK_RELEASE_REQ))
            {
              aborted = 1;
              break;
            }
          send_packet = 1;
          info->send_force_release = 0;
        }
          
      first_port += 1;
    }

  // If data to be send has been placed in the skb, terminate is and send it.
  if (send_packet)
    {
      cp = skb_put(skb, 1);
      *cp = 0;
      skb->pkt_type = PACKET_HOST;
#ifdef NSLINK_DEBUG_OUT_CMDS
      printk(KERN_INFO "nslink: packet sent (%d, %d)\n", si->out_snd_index,
	     si->in_ack_index);
#endif
      send_hdlc_packet(si, skb);
    }
  else
    {
#ifdef NSLINK_DEBUG_OUT_CMDS
      printk(KERN_INFO "nslink: Freeing unneeded skbuff\n");
#endif
      kfree_skb(skb);
    }
  if (aborted)
    goto retry;
}

/*
 *   Verify a received ACK value in HDLC mode
 */
inline static int verifyAckValue(struct si_state *si, unsigned char rxAck)
{
  struct sk_buff *ph;
  struct sk_buff *pt;
  unsigned long flags;

  //  Get pointers to the first and last skb in the retransmit queue.
  //  Must grab the list spinlock so no-one grabs the skb from under us.
  spin_lock_irqsave(&si->rtq.lock, flags);
  ph = skb_peek(&si->rtq);      // oldest packet
  pt = skb_peek_tail(&si->rtq); // newest packet

  if (ph)
    {
      unsigned char head = ((struct HDLChdr*)(ph->h.raw+3))->out_snd_index;
      unsigned char tail = pt ? ((struct HDLChdr*)(pt->h.raw+3))->out_snd_index : head;
      signed char s1 = rxAck - head;
      signed char s2 = (tail+1) - rxAck;
      if ((s1 < 0) || (s2 < 0))
        {
          printk(KERN_INFO "nslink: hdlc: [%d] bad ACK: head=%02x tail=%02x ack=%02x\n",
                 si->conc_id,head&0xff,tail&0xff,rxAck&0xff);
	  spin_unlock_irqrestore(&si->rtq.lock, flags);
          return 0;
        }
    }
  spin_unlock_irqrestore(&si->rtq.lock, flags);
  return 1;
}

/*
 * This routine processes incoming HDLC packets, and handles the
 * acknowledging of them.
 */
static void process_hdlc_packet(struct sk_buff *skb, struct si_state *si)
{
  struct HDLChdr *hdlc_hdr, *hdr2;
  unsigned long flags;
  struct sk_buff *pp;
  unsigned char ack_index, ns_save;
  unsigned char *cp;
  int len, ack_count;

  /*
   * We only process HDLC packet when we are connected, waiting
   * for a response from a connect request, or connected but
   * waiting for the response from the keepalive ping.
   */
  switch (si->state)
    {
     case SI_STATE_CONNECTING:
     case SI_STATE_CONNECTING_WAIT:
     case SI_STATE_CONNECTED:
      break;
     default:
      return;
    }

  hdlc_hdr = (struct HDLChdr *) (skb->h.raw + 3);
  skb_pull(skb, sizeof(struct HDLChdr));

  // If this is an unindexed packet, 
  if (!(hdlc_hdr->hdlc_control & CONTROL_UFRAME))
    {
#ifdef NSLINK_DEBUG_HDLC
      printk(KERN_INFO "nslink: hdlc: [%d] rx packet (%d) #%02x\n", si->conc_id,
	     hdlc_hdr->hdlc_control, hdlc_hdr->out_snd_index);
#endif

      // Check for out-of-sequence packet, discard it if so
      if (hdlc_hdr->out_snd_index != si->next_in_index)
	{
          ++si->rxBadOrderCount;
#if defined(NSLINK_DEBUG_RETRANS) || defined(NSLINK_DEBUG_HDLC)
	  printk(KERN_INFO "nslink: hdlc: [%d] rx bad seq -- got:%02x, exp:%02x, cnt=%d\n",
                 si->conc_id, hdlc_hdr->out_snd_index, si->next_in_index,si->rxBadOrderCount);
#endif
          TraceHdlc(si,5,hdlc_hdr->out_snd_index);
	  hdlc_send_ack(si, CONTROL_UFRAME);
	  return;
	}

      // Unindexed packet, in sequence, it is just an ACK.
      // If number of unacked packets exceeds window size,
      // send ack.  Otherwise, make sure rx ack timer is
      // running.
      si->rxBadOrderCount = 0;
      atomic_inc(&si->unacked_pkts);
      if (atomic_read(&si->unacked_pkts) > HDLC_MAX_UNACKED_COUNT/2)
	hdlc_send_ack(si, CONTROL_UFRAME);
      else if (!(si->hdlc_state_flags & HDLC_RCV_TIMER_ACTIVE))
	{
	  si->receiver_ack_timer = jiffies + HDLC_ACK_TIME;
	  si->hdlc_state_flags |= HDLC_RCV_TIMER_ACTIVE;
	  si_set_timer(si, HDLC_ACK_TIME);
	}
      TraceHdlc(si,2,si->next_in_index);
      si->next_in_index++;
    }
      
  // Prune retrans queue if incoming ack value changed
  if (si->in_ack_index != hdlc_hdr->next_in_index)
    {
      if (!verifyAckValue(si, hdlc_hdr->next_in_index))
        {
          TraceHdlcStop(si,-3,hdlc_hdr->next_in_index);
          goto jump;
        }
      
      TraceHdlc(si,1,hdlc_hdr->next_in_index);
      
      si->in_ack_index = hdlc_hdr->next_in_index;
      ack_index = si->in_ack_index - 1;
      ack_count = 0;
      
      /*
       * Walk through the retransmit queue, pulling off the skb pointers
       * and freeing the skb's for ack'd packets.  List lock is taken.
       *
       * Note that this assumes that we will be removing at
       * least one item from the queue.  Given that the ack
       * index has changed, this is a reasonable assumption.
       */
      spin_lock_irqsave(&si->rtq.lock, flags);
      while ((pp = __skb_dequeue(&si->rtq)))
	{
	  hdr2 = (struct HDLChdr *) (pp->h.raw + 3);
	  ns_save = hdr2->out_snd_index;

          --si->rtqCount;
	  kfree_skb(pp);
	  ack_count++;

	  if (ns_save == ack_index)
	    break;
	}
      
      si->retransCount = 0;

      // Check for an empty retransmit queue      
      if (!skb_peek(&si->rtq))
	{
#ifdef NSLINK_DEBUG_HDLC
	  printk(KERN_INFO "nslink: hdlc: [%d] all acked!\n",si->conc_id);
#endif
	  si->hdlc_state_flags &= ~HDLC_SND_TIMER_ACTIVE;
	}

      // Skb queue not empty, start retry timer
      else
	{
#ifdef NSLINK_DEBUG_HDLC
	  printk(KERN_INFO "nslink: hdlc: [%d] not all acked!\n",si->conc_id);
#endif
	  si->sender_ack_timer = jiffies + HDLC_RETRANS_TIME;
	  si->hdlc_state_flags |= HDLC_SND_TIMER_ACTIVE;
	  si_set_timer(si, HDLC_RETRANS_TIME);
	}
      spin_unlock_irqrestore(&si->rtq.lock, flags);
    }

  jump:
  
  /*
   * OK, now we actually process the data in the packet.
   *
   * XXX Currently, we assume the packet is null terminated, and
   * we don't do length checking.  We need to fix this.  Right
   * now, a malformed packet can cause the system to crash.  If
   * it's any consolation the NT driver doesn't do this sanity
   * checking either, but Unix drivers should be more robust
   * than NT drivers.  Just because it's easy to make NT machines
   * crash with malformed packets doesn't mean we have to make
   * Unix machines bug-compatible with NT.    :-)
   */
  cp = skb->data;
  len = skb->len;
  si_parse_input_pkt(si, cp, len);

  if (si->need_to_send)
    si_set_timer(si, 0);
}

//  Walk through the global array of hub config/info, searching for
//  the hub that matches the MAC address passed in.
//  Returns:  Ptr to hub config/info, NULL if not found.
static struct si_state *find_si_struct(unsigned char *eth_addr)
{
  struct si_state *si;
  int i;

  for (i = 0; i < MAX_NRP_SIS; i++)
    {
      si = conc_state[i];
      if (!si)
	continue;
      if (memcmp(si->ether_addr, eth_addr, 6) == 0)
        return si;
    }
  return 0;
}

// The networking stack calls this function to handle 0x11FE packets received.
static int comtrol_rcv(struct sk_buff *skb, 
		       struct NET_DEVICE *dev, 
		       struct packet_type *pt)
{
  struct comtrolhdr *chdr = (struct comtrolhdr *) skb->h.raw;
  struct ethhdr *ehdr = skb->mac.ethernet;
  unsigned char *hdr;
  struct adminhdr *ahdr;
  unsigned char *admin_data;
  struct si_state *si;
  struct si_box_report *rpt;

  /*
   * Drop packets not meant for us (that may show up if the
   * interface is in promiscuous mode.)
   */
  if (skb->pkt_type == PACKET_OTHERHOST)
    goto drop;

  //  Get the comtrol header from the skb, check for valid info
  hdr = skb_pull(skb, sizeof(struct comtrolhdr));
  
  if (hdr == NULL)
    goto drop;

#if 0
  printk(KERN_INFO "nslink: rx %02x,%02x  %02x:%02x:%02x:%02x:%02x:%02x\n",
         chdr->product_id,
         chdr->conc_id,
         skb->mac.ethernet->h_source[0],
         skb->mac.ethernet->h_source[1],
         skb->mac.ethernet->h_source[2],
         skb->mac.ethernet->h_source[3],
         skb->mac.ethernet->h_source[4],
         skb->mac.ethernet->h_source[5]);
#endif
  
  if (chdr->product_id != SI_ID)
    goto drop;
  
  if (chdr->conc_id >= MAX_NRP_SIS)
    goto drop;

  // Comtrol hdr OK, get hub config/info
  si = conc_state[chdr->conc_id];

  if (!si)
    goto drop;
  
  // check to see if MAC address matches the device in our hub table
  if (memcmp(skb->mac.ethernet->h_source,si->ether_addr,6))
    goto drop;
  
  switch (chdr->packet_class)
    {

    case ADMIN_CLASS:
      ahdr = (struct adminhdr *) hdr;
      admin_data = skb_pull(skb, sizeof(struct adminhdr));

      switch (ahdr->admin_type)
	{
	case ADMIN_SEC_LOAD_TYPE:
	  break;

	case ADMIN_PRD_ID_REP_TYPE:
	{
	    struct mac_id_response *idresp = (struct mac_id_response*)admin_data;

#ifdef NSLINK_DEBUG_DEVICE_ID
	    int i;
	    int len = (ahdr->len_high << 8) + ahdr->len_low;

	    printk(KERN_INFO "nslink: ID response [len=%d] device type %02x, MAC addr "
		   "%02x:%02x:%02x:%02x:%02x:%02x %s\n",
		   len,
		   chdr->product_id, admin_data[0],
		   admin_data[1], admin_data[2],
		   admin_data[3], admin_data[4],
		   admin_data[5], admin_data[6] ? "active" : "loading");

	    printk(KERN_INFO "nslink: ID response data:\n");
	    for (i = 0; i < len; ++i)
              {
                if (!(i&0x07))
                  printk(KERN_INFO "\n");
                printk(KERN_INFO "%02x ", admin_data[i]);
              }
	    printk(KERN_INFO "\n");

	    if (len >= sizeof *idresp)
	      {
		unsigned char *AppData;
		unsigned char AppDataCount;

		printk(KERN_INFO "nslink:        Model = %d\n", ntohl(idresp->ModelId));
		printk(KERN_INFO "nslink:         Arch = %d\n", idresp->ArchId);
		printk(KERN_INFO "nslink:    Num Ports = %d\n", idresp->NumPorts);
		printk(KERN_INFO "nslink:   Program Id = %d\n", ntohs(idresp->ProgramId));
		printk(KERN_INFO "nslink:    Board Rev = %02x\n",
		       idresp->BoardRev + 'A' - 1);
		printk(KERN_INFO "nslink:     Boot Rev = %d.%02d\n",
		       idresp->BootRevMajor, idresp->BootRevMinor);
		printk(KERN_INFO "nslink:     Num Apps = %d\n", idresp->NumApps);
		for (i = 0; i < idresp->NumApps; ++i)
		  {
		    printk(KERN_INFO "nslink:         App[%d] Id = %5d  ", i,
			   htons(idresp->AppRecord[i].AppId));
		    printk(KERN_INFO "version %d.%02d\n",
			   idresp->AppRecord[i].AppRevMajor,
			   idresp->AppRecord[i].AppRevMinor);
		  }

		AppData =
		  &idresp->NumApps +
		  (idresp->NumApps * sizeof idresp->AppRecord[0]) + 1;
		AppDataCount = AppData[-1];

		printk(KERN_INFO "nslink:     App Data = [%d]: ", AppDataCount);
		for (i = 0; i < AppDataCount; ++i)
		  printk(KERN_INFO "%02x ", AppData[i]);
		printk(KERN_INFO "\n");
	      }
#endif

	    /*
	     * note:  admin_data should be same as
	     * ehdr->h_source!
	     */
	    // Get ptr to hub config/info.
	    si = find_si_struct(admin_data);
	    if (si)
	    {
		if (admin_data[6])
		    si->state = SI_STATE_CONNECTING;
		else
		    si->state = SI_STATE_RESET;
	    }
	    si->modelID = ntohl(idresp->ModelId);
	    if ((num_box_rpts < MAX_NRP_SIS) &&
		(!rpt_loading_only || !admin_data[6]))
	    {
		rpt = si_box_rpts + num_box_rpts;
		memcpy(rpt->ether_addr, ehdr->h_source, 6);
		rpt->product_id = chdr->product_id;
		rpt->flags = admin_data[6];
		num_box_rpts++;
	    }
	  }
          break;   //  case ADMIN_PRD_ID_REP_TYPE:
	}          //  end of switch (ahdr->admin_type) 


      break;  // switch (chdr->packet_class) case ADMIN_CLASS:

    // "Async" class packet, process...      
    case SI_CLASS:
      if (si)
	process_hdlc_packet(skb, si);
      break;

    }

drop:
  kfree_skb(skb);
  return (0);
}

/*
 *  The timer function happens here, only used for HDLC connections.  This function 
 *  is not only the timer bottom half,  it is called by many other functions when 
 *  something needs to be sent to the hub.  The atomic timer_active variable provides 
 *  a means by which it can only be run one at a time.  If another process/thread/whatever
 *  calls this function while it is already being run, it will return immediately.  This 
 *  opens up the possible delay of a needed action, but solves the reentrancy problem.
 */
static void si_timer(unsigned long arg)
{
  struct si_state *si = (struct si_state *) arg;
  int tmp;
  unsigned long next_time = T_INFINITE;
  int i;

  if (atomic_dec_and_test(&si->timer_active) == 0)
    {
      /* The timer must be already running */
      atomic_inc(&si->timer_active);
      return;
    }

#ifdef NSLINK_DEBUG_TIMER
  printk(KERN_INFO "nslink: In si_timer (%d)...", atomic_read(&si->timer_active));
#endif
  
  if (si->state < SI_STATE_CONNECTED)
    {
      if (si->state == SI_STATE_RESET)
        {
          // signal approprate daemon that we need the box booted;
          wake_up_interruptible(&si->waitqueue);
        }
      
      if (si->state == SI_STATE_BOOTED)
	{
	  send_query_packet(si->dev, si);
	  if ((5 * HZ) < next_time)
            next_time = 5 * HZ;
	}

      /*
       * If we're in the state SI_STATE_CONNECTING, then send the
       * HDLC connect packet.  Once we receive the connect packet,
       * we will enter the SI_STATE_CONNECTED.
       */
      if (si->state == SI_STATE_CONNECTING)
	{
	  tmp = si->send_connect_time - jiffies;
	  if (tmp > 0)
	    {
	      if (tmp < next_time)
		next_time = tmp;
	    }
	  else
	    {
	      if (si->resend_counter >= 5)
		{
		  si->state = SI_STATE_RESET;
		  si->resend_counter = 0;
#ifdef NSLINK_DEBUG_TIMER
		  printk(KERN_INFO "nslink: assign timeout");
#endif
		}
	      else
		{
		  si->resend_counter++;
		  hdlc_assign_ports(si);
		  si->send_connect_time = jiffies + HDLC_RETRANS_TIME;
		}
	      if (HDLC_RETRANS_TIME < next_time)
		next_time = HDLC_RETRANS_TIME;
	    }
	}
    }
#ifdef NSLINK_DEBUG_TIMER
  printk(KERN_INFO "nslink: hdlc_state_flags=%d...", si->hdlc_state_flags);
#endif
  
  if (si->state == SI_STATE_CONNECTED && si->linkTimeoutJiffies)
    {
      /* need to send RK_CONNECT_CHECK ? */
      tmp = si->keep_alive_tx_timer - jiffies;
      if (tmp > 0)
        {
          if (tmp < next_time) 
            next_time = tmp;
        }
      else
        {
#ifdef NSLINK_DEBUG_KEEPALIVE
          printk(KERN_INFO "nslink: tx RK_CONNECT_CHECK: %d\n",si->conc_id);
#endif          
          si->need_to_send = 1;
          si->need_to_send_connect_check = 1;
          si->keep_alive_tx_timer += si->linkTimeoutJiffies/2;
          if (si->keep_alive_tx_timer - jiffies < next_time)
            next_time = si->keep_alive_tx_timer - jiffies;
        }
      
      /* link timeout? */
      tmp = si->keep_alive_timeout - jiffies;
      if (tmp > 0)
        {
          if (tmp < next_time) 
            next_time = tmp;
        }
      else 
        {
          printk(KERN_INFO "nslink: hdlc link timeout on hub %d\n",si->conc_id);
          si->state = SI_STATE_RESET;
          si->hdlc_state_flags = 0;
          if ((5 * HZ) < next_time)
            next_time = 5 * HZ;
          set_otherclose_all_tty(si);
          wake_all_tty(si);
        }
    }
  
  /*
   * Check the state again in case we declared the device as
   * dead.
   */
  if (si->state == SI_STATE_CONNECTED)
    {
      /*
       * If there are characters in the receive buffer,
       * try to send them to the line discpline.
       */
      for (i = 0; i < si->num_ports; i++)
        send_recv_buf_to_ldisc(nrp_table[i + si->base_port]);
      
      /*
       * See if we need to send commands to the remote.
       */
      if (si->need_to_send)
	{
	  int pkts_out;
	  if (si->out_snd_index >= si->in_ack_index)
            pkts_out = (int) (si->out_snd_index - si->in_ack_index);
	  else
            pkts_out = (int) (si->out_snd_index + 1 + (0xff - si->in_ack_index));

	  if (pkts_out < HDLC_MAX_UNACKED_COUNT)
            hdlc_send_remote_cmds(si);
	}
    }
  
  /*
   * If we haven't transmited any HDLC packets within the ack
   * timeout period, send a bare ack packet to prevent the
   * remote from resending packets at us. 
   */
  
  if (si->hdlc_state_flags & HDLC_RCV_TIMER_ACTIVE)
    {
#ifdef NSLINK_DEBUG_TIMER
      printk(KERN_INFO "nslink: rcv timer...");
#endif
      tmp = si->receiver_ack_timer - jiffies;
      if (tmp > 0)
	{
	  if (tmp < next_time)
	    next_time = tmp;
	}
      else
	{
#ifdef NSLINK_DEBUG_TIMER
	  printk(KERN_INFO "nslink: hdlc timeout...");
#endif
	  hdlc_send_ack(si, CONTROL_UFRAME);
	}
    }
  /*
   * If we haven't gotten an ack for the packets which we have
   * sent, resend the packets in the retransmit queue
   */
  if (si->hdlc_state_flags & HDLC_SND_TIMER_ACTIVE)
    {
#if 0
      printk(KERN_INFO "nslink: ack_timer: %ld, %ld...", si->sender_ack_timer, jiffies);
#endif
      tmp = si->sender_ack_timer - jiffies;
      if (tmp > 0)
	{
	  if (tmp < next_time)
	    next_time = tmp;
	}
      else
	{
	  hdlc_retransmit(si);
	  if (si->hdlc_state_flags & HDLC_SND_TIMER_ACTIVE)
	    {
	      si->sender_ack_timer = jiffies + HDLC_RETRANS_TIME;
	      if (HDLC_RETRANS_TIME < next_time)
		next_time = HDLC_RETRANS_TIME;
	    }
	}
    }
  if (next_time != T_INFINITE)
    si_set_timer(si, next_time);

  atomic_inc(&si->timer_active);

#ifdef NSLINK_DEBUG_TIMER
  printk(KERN_INFO "nslink: End si_timer (next_time = %d)\n", next_time);
#endif
}

static void si_set_timer(struct si_state *si, unsigned long ticks)
{
  unsigned long new_time;

#ifdef NSLINK_DEBUG_TIMER
  printk(KERN_INFO "nslink: set_timer(%d)...", ticks);
#endif
  
  if (si->use_tcp)
    return;
  
  new_time = jiffies + ticks;
  
  if ((!timer_pending(&si->timer)) ||
      ((si->timer.expires != new_time && ((si->timer.expires - jiffies) > ticks))))
    mod_timer(&si->timer,new_time);
}

static struct packet_type comtrol_packet_type = {
  .type	= __constant_htons(ETH_P_COMTROL),
  .func	= comtrol_rcv,
};

/********************************************************************************/
/*                                                                              */
/*        MAC / TCP COMMON FUNCTIONS                                            */
/*                                                                              */
/********************************************************************************/

/*
 *  Process the data portion of an incoming (HDLC or TCP) packet.  We do this by
 *  walking through the received data buffer, processing the commands until the terminating
 *  byte is reached.  Note that every command has a length (cmd + associated data).  If the 
 *  packet length is too short for the command, this is a fragment (caused by tcp).  Keep
 *  this fragment in a different buffer for processing when the next packet is received.
 */
static void si_parse_input_pkt(struct si_state *si, 
			       unsigned char *buf, 
			       unsigned long len)
{
  int frcnt, portnum=-1;
  unsigned int cmdval;
  struct nr_port *info;
  struct si_port_state *pstate;
  unsigned long flags;

  info = NULL;
  pstate = NULL;
  
  // might have fragmented a packet so we need to remember our
  // port number from last time
  if (si->use_tcp)
    {
      portnum = si->tcp_last_port;
      if ((portnum >= 0) && (portnum < si->num_ports))
	{
	  info = nrp_table[portnum + si->base_port];
	  pstate = si->port_state + portnum;
	}
    }

  //printk(KERN_INFO "nslink: parse_input_packet(%p,%p,%d)\n",si,buf,len);
  //printk(KERN_INFO "nslink: info=%p pstate=%p\n",info,pstate);
  
  while ((len > 0) && (*buf))
    {
      switch (*buf)
	{
	case RK_CONNECT_CHECK:
#if defined(NSLINK_DEBUG_IN_CMDS) || defined(NSLINK_DEBUG_KEEPALIVE)
	  printk(KERN_INFO "nslink: rx RK_CONNECT_CHECK: %d\n",si->conc_id);
#endif
          if (si->linkTimeoutJiffies)
            {
              si->keep_alive_timeout = jiffies + si->linkTimeoutJiffies;
              if (!si->use_tcp)
                si_set_timer(si, si->linkTimeoutJiffies);
            }
	  frcnt = 1;
	  break;
	case RK_PORT_SET:
	  if (len < 2)
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;
	    }
	  portnum = buf[1];
	  /* printk(KERN_INFO "nslink: RK_PORT_SET: %d\n",portnum); */
	  if ((portnum >= 0) && (portnum < si->num_ports))
	    {
	      si->tcp_last_port = portnum;
	      info = nrp_table[portnum + si->base_port];
	      pstate = si->port_state + portnum;
	      frcnt = 2;
	    }
	  else
	    {
	      printk(KERN_INFO "nslink: RK_PORT_SET: invalid port %d\n", portnum);
	      info = NULL;
	      pstate = NULL;
	      frcnt = len;
	    }
	  break;
	case RK_MSR_SET:
	  if (len < 3)
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;
	    }
	  cmdval = buf[1];
	  cmdval += buf[2] << 8;

//	  printk(KERN_INFO "nslink: RK_MSR_SET: 0x%x\n",cmdval); 

	  if (pstate)
            pstate->modemStatusRegister = cmdval;
	  if (info && info->tty && (info->flags & SI_INITIALIZED))
	    {
	      if ((cmdval & MSR_CD_ON) && (!info->cd_status))
		{
		  wake_up_interruptible(&info->open_wait);
		  info->cd_status = 1;
		}
	      else if (!(cmdval & MSR_CD_ON) && (info->cd_status))
                {
                  if (!C_CLOCAL(info->tty))
                    tty_hangup(info->tty);
                  info->cd_status = 0;
		}
	    }
	  frcnt = 3;
	  break;
	case RK_ESR_SET:
	  if (len < 3)
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;
	    }
	  cmdval = buf[1];
	  cmdval += buf[2] << 8;
	  /* printk(KERN_INFO "nslink: RK_ESR_SET: 0x%x\n",cmdval); */
	  if (pstate)
	    {
	      pstate->errorStatusRegister = cmdval;

	      if (pstate->errorStatusRegister & ESR_FRAME_ERROR)
		  info->framingErrors++;
	      if (pstate->errorStatusRegister & ESR_PARITY_ERROR)
		  info->parityErrors++;
	      if (pstate->errorStatusRegister & ESR_OVERFLOW_ERROR)
		  info->overflowErrors++;

	    }
	  frcnt = 3;
	  break;
	case RK_ACTION_ACK:
	  if (len < 3)
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;                         
	    }
	  cmdval = buf[1];
	  cmdval += buf[2] << 8;
	  /* printk(KERN_INFO "nslink: RK_ACTION_ACK: 0x%x\n",cmdval); */
	  if (pstate)
            pstate->actionResponse = cmdval;
	  frcnt = 3;
	  break;
	case RK_QIN_STATUS:
	  if (len < 3)
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;
	    }
	  cmdval = buf[1];
	  cmdval += buf[2] << 8;
	  /* printk(KERN_INFO "nslink: RK_QIN_STATUS: 0x%x\n",cmdval); */
	  if (pstate)
	    {
	      pstate->nGetRemote = cmdval;
              TraceQueue(info,4,cmdval);
	      if (info && info->tty && 
                  (info->flags & SI_INITIALIZED) && 
		  info->xmit_cnt &&
                  get_tx_space_remote(pstate))
                si->need_to_send = 1;
	    }
	  frcnt = 3;
	  break;                              
	case RK_DATA_BLK:
	  if (len < 3)
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;
	    }
	  cmdval = buf[1];
	  cmdval += buf[2] << 8;
	  if (len < (cmdval + 3))
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;
	    }
	  if (info &&
	      info->tty &&
	      !(info->flags & SI_CLOSING) &&
	      (info->flags & SI_INITIALIZED))
	    {
              info->rxBytes += cmdval;
	      if (info->recv_cnt || (cmdval >= (unsigned int)info->tty->ldisc.receive_room(info->tty)))
                push_chars_to_recv_buf(info, &buf[3], cmdval);
	      else
		{
		  info->tty->ldisc.receive_buf(info->tty, &buf[3],NULL,cmdval);

		  spin_lock_irqsave(&pstate->nGetLocalLock, flags);
		  pstate->nGetLocal += cmdval;
		  spin_unlock_irqrestore(&pstate->nGetLocalLock, flags);

		  pstate->update_queue = 1;
		  si->need_to_send = 1;
		}
	    }
	  else
	    {
	      if (pstate)
		{
		  spin_lock_irqsave(&pstate->nGetLocalLock, flags);
		  pstate->nGetLocal += cmdval;
		  spin_unlock_irqrestore(&pstate->nGetLocalLock, flags);

		  pstate->update_queue = 1;
		  si->need_to_send = 1;
		}
	    }
	  frcnt = 3 + cmdval;
	  break;
	case RK_ASSIGN_REPLY:
	  if (si->state == SI_STATE_CONNECTING_WAIT ||
              si->state == SI_STATE_CONNECTING)
	    {
	      printk(KERN_INFO "nslink: Connected -- %d.\n", si->base_port);
	      si->state = SI_STATE_CONNECTED;
              si->update_timeout = 1;
              si->need_to_send = 1;
#ifdef NSLINK_DEBUG_KEEPALIVE
              printk(KERN_INFO "nslink: init link timer %d: %ld jiffies\n",si->conc_id,si->linkTimeoutJiffies/2);
#endif              
              if (si->linkTimeoutJiffies)
                {
                  si->keep_alive_tx_timer = jiffies + si->linkTimeoutJiffies/2;
                  si->keep_alive_timeout  = jiffies + si->linkTimeoutJiffies;
                  if (!si->use_tcp)
                    si_set_timer(si, si->linkTimeoutJiffies/2);
                }
              clr_otherclose_all_tty(si);
	      port_resync_all(si);
	    }
	  else
            printk(KERN_INFO "nslink: Unexpected RPSH-Si connection assign reply!\n");
	  frcnt = 1;
	  break;
	case RK_IC_TIMEOUT:
//	  printk(KERN_INFO "nslink: RK_IC_TIMEOUT received, setting MSR_ICTO_OCCURRED \n"); 
	  info->si->port_state[info->portnum].modemStatusRegister |= MSR_ICTO_OCCURRED;
	  frcnt = 1;
	  break;
	case RK_REMOTE_STATUS:
	  if (len < 2)
	    {
	      if (si->use_tcp)
                tcp_store_fragment(si, buf, len);
	      return;
	    }
	  if (buf[1] == REM_PORT_ASSIGNED)
	    {
	      /* assign ok */
	      info->remote_status = REM_PORT_ASSIGNED;
	      cmdval = buf[2];
	      cmdval += buf[3] << 8;

//	      printk(KERN_INFO "nslink: RK_REMOTE_STATUS: 0x%x\n",cmdval); 

	      if (pstate)
                pstate->modemStatusRegister = cmdval;
	      if (info && info->tty && (info->flags & SI_INITIALIZED))
		{
		  if ((cmdval & MSR_CD_ON) && (!info->cd_status))
		    {
		      wake_up_interruptible(&info->open_wait);
		      info->cd_status = 1;
		    }
		  else if (!(cmdval & MSR_CD_ON) && (info->cd_status))
                    {
                      if (!C_CLOCAL(info->tty))
                        tty_hangup(info->tty);
                      info->cd_status = 0;
		    }
		}
	      frcnt = 4;
	    }
	  else
            {
              /* port not available */
              //printk(KERN_INFO "nslink: port %d unavailable\n",portnum);
              //printk(KERN_INFO "nslink: info = %p\n",info);
              //printk(KERN_INFO "nslink: pstate = %p\n",pstate);
              if (info)
                {
                  int old_status = info->remote_status;
                  info->conn_refused = 1;
                  
                  if (buf[1] == 0x11)
                    info->remote_status = REM_PORT_BUSY;
                  else
                    {
                      info->remote_status = REM_PORT_AVAIL;
                      info->force_release_acked = 1;
                      wake_up_interruptible(&info->release_wait);
                    }
                  
                  if (old_status != REM_PORT_BUSY && old_status != REM_PORT_AVAIL)
                    {
                      if (info->tty)
                        {
                          set_bit(TTY_OTHER_CLOSED, &info->tty->flags);
                          wake_tty(info->tty);
                        }
                      port_resync(info, pstate);
                    }
                }
              frcnt = 2;
            }
	  break;
	default:
	  printk(KERN_INFO "nslink: ParseInputPkt invalid cmd 0x%x, len %ld\n", *buf, len);
	  frcnt = len;
	  break;
	}
      len -= frcnt;
      buf += frcnt;
    }

  return;
}

static int init_hub_state(struct si_box_config *in, int from_user)
{
  struct si_state *si;
  struct NET_DEVICE *dev;
  int i, ret;
  unsigned int size;
  struct si_box_config *cfg, tmp;

  
  if (from_user)
    {
      if (copy_from_user(&tmp, (void *) in, sizeof(struct si_box_config)))
	return -EFAULT;
      cfg = &tmp;
    }
  else
    cfg = in;

  if (cfg->boxid >= MAX_NRP_SIS)
    return -EINVAL;

  if (cfg->ip_addr)
    dev = NULL;
  else
    {
      dev = dev_get_by_name(cfg->interface);
      if (!dev)
        {
          printk(KERN_INFO "nslink: can't find network device '%s'\n",cfg->interface);
          return -ENOENT;
        }
    }

  if (conc_state[cfg->boxid])
    {
      ret = free_si_box(cfg->boxid);
      if (ret)
	return ret;
      conc_state[cfg->boxid] = 0;
    }
  if (cfg->num_ports == 0)
    return 0;

  si = kmalloc(sizeof(struct si_state), GFP_KERNEL);
  if (!si)
    return -ENOMEM;
  memset(si, 0, sizeof(struct si_state));

  init_timer(&si->timer);
  memcpy(si->ether_addr, cfg->ether_addr, 6);
  si->dev = dev;
  skb_queue_head_init(&si->rtq);
  si->conc_id = cfg->boxid;
  si->base_port = cfg->base_port;
  si->num_ports = cfg->num_ports;
  memcpy(si->rsmode, cfg->rsmode, MAX_NRP_PORTS);
  si->link_timeout = cfg->link_timeout;
  si->scan_period = cfg->scan_period;
  si->arch_id = cfg->arch_id;
  si->linkTimeoutJiffies = si->link_timeout * HZ;

#ifdef NSLINK_DEBUG_KEEPALIVE
  printk(KERN_INFO "nslink: linkTimeout for hub %d = %d seconds\n",si->conc_id,si->link_timeout);
#endif  
  si->ip_addr = cfg->ip_addr;
  if (si->ip_addr)
    si->use_tcp = 1;
  else
    si->use_tcp = 0;
  si->timer.data = (unsigned long) si;
  si->timer.function = si_timer;
#ifdef atomic_set
  atomic_set(&si->timer_active, 1);
#else
  si->timer_active = 1;
#endif

  size = cfg->num_ports * sizeof(struct si_port_state);
  si->port_state = kmalloc(size, GFP_KERNEL);
  if (!si->port_state)
    {
      kfree(si);
      return -ENOMEM;
    }
  memset(si->port_state, 0, size);

  init_waitqueue_head(&si->waitqueue);

  for (i = 0; i < cfg->num_ports; i++)
    init_nr_port(si, i, cfg->inter_char_timeout[i], cfg->low_latency[i]);

  conc_state[cfg->boxid] = si;

  if (si->dev)
    {
      /* do not set timer yet if there is not device, i.e., IP addressing. */
      si_set_timer(si, HZ);
    }

  return 0;
}

/*
 *  This function initializes the nr_port structure for one serial port.
 *  Inputs:  si-> ptr to si struct for the hub, portnum (0-31) relative port number
 *           (not the ttySIx number)
 */
static void init_nr_port(struct si_state *si, int portnum, int ic_timeout, int low_latency)
{
  struct nr_port *info;
  struct si_port_state *pstate;
  int line;

  line = si->base_port + portnum;
  pstate = si->port_state + portnum;

  info = kmalloc(sizeof(struct nr_port), GFP_KERNEL);
  if (!info)
    {
      printk(KERN_INFO "nslink: Couldn't allocate info struct for line #%d\n", line);
      return;
    }
  memset(info, 0, sizeof(struct nr_port));

  info->magic = RPORT_MAGIC;
  info->line = line;
  info->si = si;
  info->portnum = portnum;
  info->closing_wait = 3000;
  info->close_delay = 50;
  info->baud_rate = 9600;
  info->low_latency = low_latency;
  info->inter_char_timeout = ic_timeout;
  info->normal_termios = nslink_driver.init_termios;
  init_waitqueue_head(&info->open_wait);
  init_waitqueue_head(&info->close_wait);
  init_waitqueue_head(&info->release_wait);

  spin_lock_init(&info->slock);
  spin_lock_init(&pstate->nGetLocalLock);
  sema_init(&info->write_sem, 1);
  nrp_table[line] = info;

#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)
  tty_register_device(&nslink_driver, line, NULL);
#endif

}


/********************************************************************************/
/*                                                                              */
/*        TTY FUNCTIONS                                                         */
/*                                                                              */
/********************************************************************************/

/*
 * Utility function.
 */
#ifndef MIN
#define MIN(a,b)	((a) < (b) ? (a) : (b))
#endif

/*
 * If there is characters in the nr_port recv_buffer, try to send it out to
 * the tty line discpline.
 */
static void send_recv_buf_to_ldisc(struct nr_port *info)
{
  struct tty_struct *tty;
  int c, total_sent = 0;
  struct si_port_state *pstate;
  unsigned long flags;

  if (!info || !info->tty || !info->recv_cnt || !info->recv_buf)
    return;

  tty = info->tty;

  //  Copy the data into tty bufs, spinlock to protect recv_tail and recv_cnt
  spin_lock_irqsave(&info->slock, flags);
  while (1)
    {
      c = tty->ldisc.receive_room(tty);
      if (c > info->recv_cnt)
	  c = info->recv_cnt;

      if (c > (RECV_BUF_SIZE - info->recv_tail))
	  c = RECV_BUF_SIZE - info->recv_tail;

      if (c <= 0)
	break;
      tty->ldisc.receive_buf(tty, info->recv_buf + info->recv_tail, NULL, c);
      info->recv_tail += c;
      info->recv_tail &= RECV_BUF_SIZE - 1;
      info->recv_cnt -= c;
      total_sent += c;
    }
  spin_unlock_irqrestore(&info->slock, flags);

  if (total_sent)
    {
      pstate = info->si->port_state + info->portnum;

      spin_lock_irqsave(&pstate->nGetLocalLock, flags);
      pstate->nGetLocal += total_sent;
      spin_unlock_irqrestore(&pstate->nGetLocalLock, flags);

      pstate->update_queue = 1;
      info->si->need_to_send = 1;
      /*
       * We don't need to set the timer since we're called
       * by si_timer(), and the code to send the remote
       * cmds happens immediately after the calll to this
       * function.
       */
    }
  /*
   * If there are still more characters to send, try again in
   * 0.5 seconds later.
   */
  if (info->recv_cnt)
    si_set_timer(info->si, HZ / 2);
}

/*
 *  Push characters to the nr_port recv_buffer.  Data is from a received
 *  ethernet packet (either MAC mode or TCP).
 */
static void push_chars_to_recv_buf(struct nr_port *info, unsigned char *cp, int len)
{
  struct tty_struct *tty;
  int c;
  unsigned long flags;

  if (!info || !info->tty || !info->recv_buf)
    return;

  tty = info->tty;

#if 0
  printk(KERN_INFO "nslink: pushed %d chars to recv_buf...", len);
#endif

  //  Copy the data into port buf, spinlock to protect recv_head and xmit_cnt
  spin_lock_irqsave(&info->slock, flags);
  while (1)
    {
      c = RECV_BUF_SIZE - info->recv_cnt - 1;
      if (c > len)
	c = len;
      if (c > RECV_BUF_SIZE - info->recv_head)
	c = RECV_BUF_SIZE - info->recv_head;
      if (c <= 0)
	break;
      memcpy(info->recv_buf + info->recv_head, cp, c);
      info->recv_head = (info->recv_head + c) & (RECV_BUF_SIZE - 1);
      info->recv_cnt += c;
      cp += c;
      len -= c;
    }
  spin_unlock_irqrestore(&info->slock, flags);

  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);
}

static inline int nslink_paranoia_check(struct nr_port *info, const char *routine)
{
#ifdef NSLINK_PARANOIA_CHECK
  if (!info)
    return 1;
  if (info->magic != RPORT_MAGIC) {
      printk(KERN_INFO "Warning: bad magic number for Nslink struct in %s\n",
	     routine);
      return 1;
    }
#endif
  return 0;
}

static int baud_table[] = 
{
  0, 50, 75, 110, 134, 150, 200, 300,
  600, 1200, 1800, 2400, 4800, 9600, 19200,
  38400, 57600, 115200, 230400, 460800, 0
};

/*
 * This routine configures an nslink port according to its termio settings.
 */
static void configure_nr_port(struct nr_port *info)
{
  unsigned cflag;
  int i, bits, baud;
  int control_settings = 0;

  if (!info->tty || !info->tty->termios)
    return;
  cflag = info->tty->termios->c_cflag;

  /* Byte size and parity */
  if ((cflag & CSIZE) == CS8)
    {
      control_settings |= SC_DATABITS_8;
      bits = 10;
    }
  else
    {
      control_settings |= SC_DATABITS_7;
      bits = 9;
    }
  if (cflag & CSTOPB)
    {
      control_settings |= SC_STOPBITS_2;
      bits++;
    }
  else
    control_settings |= SC_STOPBITS_1;

  if (cflag & PARENB)
    {
      bits++;
      if (cflag & PARODD)
	control_settings |= SC_PARITY_ODD;
      else
	control_settings |= SC_PARITY_EVEN;
    }
  
  if (cflag & CRTSCTS)
    control_settings |= SC_FLOW_CTS_AUTO+SC_FLOW_RTS_AUTO;
  if (info->si->rsmode[info->portnum] == 0x85)
    control_settings |= SC_FLOW_RTS_ARS485;
  
  if (I_IXON(info->tty))
    control_settings |= SC_FLOW_XON_TX_AUTO;

  if (I_IXOFF(info->tty))
    control_settings |= SC_FLOW_XON_RX_AUTO;

  if (control_settings != info->control_settings)
    {
      info->control_settings = control_settings;
      info->update_control_settings = 1;
    }

  /* baud rate */
  i = cflag & CBAUD;
  if (i & CBAUDEX)
    {
      i &= ~CBAUDEX;
      if (i < 1 || i > 4)
	info->tty->termios->c_cflag &= ~CBAUDEX;
      else
	i += 15;
    }
  if (i == 15)
    {
      if ((info->flags & SI_SPD_MASK) == SI_SPD_HI)
	i += 1;
      if ((info->flags & SI_SPD_MASK) == SI_SPD_VHI)
	i += 2;
      if ((info->flags & SI_SPD_MASK) == SI_SPD_SHI)
	i += 3;
      if ((info->flags & SI_SPD_MASK) == SI_SPD_WARP)
	i += 4;
    }
  baud = baud_table[i] ? baud_table[i] : 9600;
  info->cps = baud / bits;

  if (baud != info->baud_rate)
    {
      info->baud_rate = baud;
      info->update_baud_rate = 1;
    }

  if (info->rx_xon != START_CHAR(info->tty) ||
      info->rx_xoff != STOP_CHAR(info->tty) ||
      info->tx_xon != START_CHAR(info->tty) ||
      info->tx_xoff != STOP_CHAR(info->tty))
    {
      info->rx_xon = info->tx_xon = START_CHAR(info->tty);
      info->rx_xoff = info->tx_xoff = STOP_CHAR(info->tty);
      info->update_special_charset = 1;
    }
      
  
  if (info->update_baud_rate || info->update_control_settings)
    {
      info->si->need_to_send = 1;
      if (info->si->use_tcp == 0)
        si_timer((unsigned long) info->si);
    }
}

static int block_til_ready(struct tty_struct *tty, 
			   struct file *filp, 
			   struct nr_port *info)
{
  wait_queue_t wait;
  int retval;
  int do_clocal = 0, extra_count = 0;
  unsigned long flags;

  init_waitqueue_entry(&wait,current);
  
  /*
   * If the device is in the middle of being closed, then block
   * until it's done, and then try again.
   */
  if (tty_hung_up_p(filp))
    return ((info->flags & SI_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);

  // If we are in the process of closing, wait for it to complete, then return
  if (info->flags & SI_CLOSING)
    {
      interruptible_sleep_on(&info->close_wait);
      return ((info->flags & SI_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

  /*
   * If non-blocking mode is set, or the port is not enabled,
   * then make the check up front and then exit.
   */
  if ((filp->f_flags & O_NONBLOCK) || (tty->flags & (1 << TTY_IO_ERROR)))
    {
      info->flags |= SI_NORMAL_ACTIVE;
      return 0;
    }

  if (tty->termios->c_cflag & CLOCAL)
    do_clocal = 1;

  /*
   * Block waiting for the carrier detect and the line to become
   * free.  While we are in this loop, info->count is dropped by one, so that
   * nrp_close() knows when to free things.  We restore it upon
   * exit, either normal or abnormal.
   */
  retval = 0;
  add_wait_queue(&info->open_wait, &wait);
#ifdef NSLINK_DEBUG_OPEN
  printk(KERN_INFO "nslink: block_til_ready before block: ttySI%d, count = %d\n",
	 info->line, info->count);
#endif

  //  Use spinlock to protect info->count
  spin_lock_irqsave(&info->slock, flags);
  if (!tty_hung_up_p(filp))
    {
      extra_count = 1;
      info->count--;
    }
  spin_unlock_irqrestore(&info->slock, flags);

  info->blocked_open++;

  while (1)
    {
      if (tty->termios->c_cflag & CBAUD)
	{
	  if ((info->si->rsmode[info->portnum] != 0x85) &&
	      !(info->modem_control_register & MCR_DTR_SET_MASK))
	    {
	      info->modem_control_register |= MCR_DTR_SET_ON | MCR_RTS_SET_ON;
	      info->update_mcr = 1;
	      info->si->need_to_send = 1;
	      if (info->si->use_tcp == 0)
                si_timer((unsigned long) info->si);
	    }
	}
      set_current_state(TASK_INTERRUPTIBLE);
      if (tty_hung_up_p(filp) || !(info->flags & SI_INITIALIZED))
	{
	  if (info->flags & SI_HUP_NOTIFY)
	    retval = -EAGAIN;
	  else
	    retval = -ERESTARTSYS;
	  break;
	}

      // kdr  This appears to break always...
#if 0
      if (!(info->flags & SI_CALLOUT_ACTIVE) &&
	  !(info->flags & SI_CLOSING) &&
	  (do_clocal || (sGetChanStatusLo(&info->channel) & CD_ACT)))
	break;
#else
      break;
#endif
      if (signal_pending(current))
	{
	  retval = -ERESTARTSYS;
	  break;
	}
#ifdef NSLINK_DEBUG_OPEN
      printk(KERN_INFO "nslink: block_til_ready blocking: ttySI%d, count = %d, flags=0x%0x\n",
	 info->line, info->count, info->flags);
#endif
      schedule();
    }
  set_current_state(TASK_RUNNING);
  remove_wait_queue(&info->open_wait, &wait);

  //  Use spinlock to protect info->count
  spin_lock_irqsave(&info->slock, flags);
  if (extra_count)
    info->count++;
  spin_unlock_irqrestore(&info->slock, flags);

  info->blocked_open--;
#ifdef NSLINK_DEBUG_OPEN
  printk(KERN_INFO "nslink: block_til_ready after blocking: ttySI%d, count = %d\n",
	 info->line, info->count);
#endif
  if (retval)
    return retval;
  info->flags |= SI_NORMAL_ACTIVE;
  return 0;
}

/*  
 *  TTY port open handler
 */
static int nrp_open(struct tty_struct *tty, struct file *filp)
{
  struct si_port_state *pstate;
  struct nr_port *info;
  int line, retval, timeout;
  unsigned long page, page2, jifval;
  unsigned long flags;

  // 'line' is the port number 'xx' /dev/ttySIxx
  line = TTY_GET_LINE(tty);
  
  if ((line < 0) || (line >= (MAX_NRP_SIS * MAX_NRP_PORTS)))
    return -ENODEV;
    
  // 'info' is ptr to global port config/info data
  info = nrp_table[line];
  
  if (!info || !info->si || info->si->rsmode[info->portnum] == 0xde)
    return -ENODEV;

  page = __get_free_page(GFP_KERNEL);
  if (!page)
    return -ENOMEM;

  page2 = __get_free_page(GFP_KERNEL);
  if (!page2)
    {
      free_page(page);
      return -ENOMEM;
    }

  info->remote_status = REM_OPEN_PORT;
  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);

  //  Wait a sec (less time if signalled)
  jifval = jiffies + (HZ);
  while ((jifval > jiffies) && (info->remote_status & (REM_OPEN_PORT | REM_WAITING_OPEN)))
    {
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout((HZ / 10));
      if (signal_pending(current))
	break;
    }

  if (info->remote_status != REM_PORT_ASSIGNED)
    {
      free_page(page2);
      free_page(page);
      return -EBUSY;
    }
  
  tty->driver_data = info;

  if (info == NULL)
    {
      tty->flags = (1 << TTY_IO_ERROR);
      free_page(page2);
      free_page(page);
      return 0;
    }

  // If we are in the process of closing, wait for it to complete, then return
  if (info->flags & SI_CLOSING)
    {
      interruptible_sleep_on(&info->close_wait);
      free_page(page2);
      free_page(page);
      return ((info->flags & SI_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

  /*
   * We must not sleep from here until the port is marked fully
   * in use.
   */
  if (info->xmit_buf)
    free_page(page);
  else
    info->xmit_buf = (unsigned char *) page;
  if (info->recv_buf)
    free_page(page2);
  else
    {
      info->recv_buf = (unsigned char *) page2;
      info->recv_cnt = 0;
      info->recv_head = 0;
      info->recv_tail = 0;
    }
  info->tty = tty;

  // If we are in the process of closing, wait for it to complete, then return
  if (info->flags & SI_CLOSING)
    {
      interruptible_sleep_on(&info->close_wait);
      return ((info->flags & SI_HUP_NOTIFY) ? -EAGAIN : -ERESTARTSYS);
    }

  //  Use spinlock to protect info->count
  spin_lock_irqsave(&info->slock, flags);
  if (info->count++ == 0)
    {

#if ((LINUX_VERSION_CODE < VERSION_CODE(2,3,0)) && defined(MODULE))
	MOD_INC_USE_COUNT;
#endif

      atomic_inc(&nrp_num_ports_open);

#ifdef NSLINK_DEBUG_OPEN
      printk(KERN_INFO "nslink: nslink mod++ = %d...", atomic_read(&nrp_num_ports_open));
#endif

    }
    spin_unlock_irqrestore(&info->slock, flags);

#ifdef NSLINK_DEBUG_OPEN
  printk(KERN_INFO "nslink: nrp_open ttySI%d, count=%d\n", info->line, info->count);
#endif

  /*
   * Info->count is now 1; so it's safe to sleep now.
   */
  pstate = info->si->port_state + info->portnum;
  if (pstate->modemStatusRegister & MSR_CD_ON)
    info->cd_status = 1;
  else
    info->cd_status = 0;

  /*
   * XXX Shouldn't be needed, but embedded box sometimes drops a
   * transmitted data subpacket request (while still ack'ing the
   * packet at the HDLC level!)  We resync on an initial open to
   * make sure things are sane...
   */
  timeout = get_tx_cnt_remote(pstate);
  if (timeout)
    {
      timeout = (get_tx_cnt_remote(pstate) + 1) * HZ / 1200;
      if (timeout < HZ)
	timeout = HZ;
      nrp_wait_until_sent(tty, timeout);
    }

  TraceQueue(info,3,0);

  pstate->actionResponse = 0;
  info->action_register = ACT_FLUSH_INPUT | ACT_FLUSH_OUTPUT;
  info->update_action = 1;

  if (info->si->arch_id == 3 || info->si->arch_id == 2)
    info->update_ic_timeout = 1;

  /*
   * Set up the tty->alt_speed kludge
   */
  if ((info->flags & SI_SPD_MASK) == SI_SPD_HI)
    info->tty->alt_speed = 57600;
  if ((info->flags & SI_SPD_MASK) == SI_SPD_VHI)
    info->tty->alt_speed = 115200;
  if ((info->flags & SI_SPD_MASK) == SI_SPD_SHI)
    info->tty->alt_speed = 230400;
  if ((info->flags & SI_SPD_MASK) == SI_SPD_WARP)
    info->tty->alt_speed = 460800;

  configure_nr_port(info);
  
  if (info->si->rsmode[info->portnum] != 0x85)
    {
      if (tty->termios->c_cflag & CBAUD)
	info->modem_control_register |= MCR_DTR_SET_ON;
      info->modem_control_register |= MCR_RTS_SET_ON;
      info->update_mcr = 1;
    }
  
  info->si->need_to_send = 1;
  
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);
  
  // block until action response is received

  //  Wait a sec (less time if signalled)
  jifval = jiffies + (HZ);
  while ((jifval > jiffies) && (pstate->actionResponse == 0))
    {
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout((HZ / 10));
      if (signal_pending(current))
        break;
    }

  info->flags |= SI_INITIALIZED;

  retval = block_til_ready(tty, filp, info);
  if (retval)
    {
#ifdef NSLINK_DEBUG_OPEN
      printk(KERN_INFO "nslink: nrp_open returning after block_til_ready with %d\n",
	     retval);
#endif
      return retval;
    }

  if ((info->count == 1) && (info->flags & SI_SPLIT_TERMIOS))
    {
      *tty->termios = info->normal_termios;
      configure_nr_port(info);
    }
  
  tty->low_latency = info->low_latency;

  return 0;
}

/*  
 *  TTY port close handler
 */
static void nrp_close(struct tty_struct *tty, struct file *filp)
{
  struct si_port_state *pstate;
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  unsigned long flags, jifval;
  int timeout;

  if (nslink_paranoia_check(info, "nrp_close"))
    return;

  if (!info)
    return;

  if (tty_hung_up_p(filp))
      return;

#ifdef NSLINK_DEBUG_OPEN
  printk(KERN_INFO "nslink: nrp_close ttySI%d, count = %d\n", info->line, info->count);
#endif

  //  Use spinlock to protect info->count
  spin_lock_irqsave(&info->slock, flags);

  if ((tty->count == 1) && (info->count != 1))
    {
      /*
       * Uh, oh.  tty->count is 1, which means that the tty
       * structure will be freed.  Info->count should always
       * be one in these conditions.  If it's greater than
       * one, we've got real problems, since it means the
       * serial port won't be shutdown.
       */
      printk(KERN_INFO "nslink: nrp_close: bad serial port count; tty->count is 1, "
	     "info->count is %d\n", info->count);
      info->count = 1;
    }
  if (--info->count < 0)
    {
      printk(KERN_INFO "nslink: nrp_close: bad serial port count for ttySI%d: %d\n",
	     info->line, info->count);
      info->count = 0;
    }
  spin_unlock_irqrestore(&info->slock, flags);

  if (info->count) 
    return;

  info->flags |= SI_CLOSING;

  info->normal_termios = *tty->termios;

  /*
   * Notify the line discpline to only process XON/XOFF characters
   */
  tty->closing = 1;

  /*
   * If transmission was throttled by the application request,
   * just flush the xmit buffer.
   */
  if (tty->flow_stopped)
    nrp_flush_buffer(tty);

  /*
   * Wait for the transmit buffer to clear
   */
  if (info->closing_wait != SI_CLOSING_WAIT_NONE)
    tty_wait_until_sent(tty, info->closing_wait);
  /*
   * Before we drop DTR, make sure the UART transmitter
   * has completely drained; this is especially
   * important if there is a transmit FIFO!
   */
  pstate = info->si->port_state + info->portnum;
  if (info->cps == 0)
    timeout = 1;
  else
    {
      timeout = (get_tx_cnt_remote(pstate) + 1) * HZ / info->cps;
      if (timeout == 0)
        timeout = 1;
    }

  nrp_wait_until_sent(tty, timeout);

  info->action_register = ACT_FLUSH_INPUT | ACT_FLUSH_OUTPUT;
  info->update_action = 1;

  info->modem_control_register &= ~MCR_RTS_SET_MASK;
  if (C_HUPCL(tty) && info->modem_control_register & MCR_DTR_SET_MASK)
    info->modem_control_register &= ~MCR_DTR_SET_MASK;
  info->update_mcr = 1;

  info->si->need_to_send = 1;
  
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);
  else
    tcp_send_remote_cmds(info->si);

  if (TTY_DRIVER_FLUSH_BUFFER_EXISTS(tty))
    TTY_DRIVER_FLUSH_BUFFER(tty);
  if (tty->ldisc.flush_buffer)
    tty->ldisc.flush_buffer(tty);

  if (info->blocked_open)
    {
      if (info->close_delay)
	{
	  set_current_state(TASK_INTERRUPTIBLE);
	  schedule_timeout(info->close_delay);
	}
      wake_up_interruptible(&info->open_wait);
    }
  else
    {
      if (info->xmit_buf)
	{
	  free_page((unsigned long) info->xmit_buf);
	  info->xmit_buf = 0;
	}
      if (info->recv_buf)
	{
	  free_page((unsigned long) info->recv_buf);
	  info->recv_buf = 0;
	  info->recv_cnt = 0;
	}
    }
  info->flags &= ~(SI_INITIALIZED | SI_CLOSING | SI_NORMAL_ACTIVE);
  tty->closing = 0;

  // Wake up other processes waiting for the close to complete.
  wake_up_interruptible(&info->close_wait);

#if ((LINUX_VERSION_CODE < VERSION_CODE(2,3,0)) && defined(MODULE))
  MOD_DEC_USE_COUNT;
#endif

  atomic_dec(&nrp_num_ports_open);

#ifdef NSLINK_DEBUG_OPEN
  printk(KERN_INFO "nslink: nslink mod-- = %d...", atomic_read(&nrp_num_ports_open));
#endif

  info->remote_status = REM_RELEASE_PORT;
  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);

  //  Wait a sec (less time if signalled)
  jifval = jiffies + (HZ);
  while ((jifval > jiffies) && (info->remote_status & (REM_RELEASE_PORT | REM_WAITING_CLOSE))) {
    set_current_state(TASK_INTERRUPTIBLE);
    schedule_timeout((HZ / 10));
    if (signal_pending(current))
      break;
  }
  
  info->tty = NULL;
#ifdef NSLINK_DEBUG_OPEN
  printk(KERN_INFO "nslink: nrp_close ttySI%d complete shutdown\n", info->line);
#endif

}

static void nrp_set_termios(struct tty_struct *tty, struct termios *old_termios)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  unsigned cflag;
  unsigned iflag;

  if (nslink_paranoia_check(info, "nrp_set_termios"))
    return;

  cflag = tty->termios->c_cflag;
  iflag = tty->termios->c_iflag;

  if (cflag == old_termios->c_cflag && iflag == old_termios->c_iflag)
    return;

  /*
   * This driver doesn't support CS5 or CS6
   */
  if (((cflag & CSIZE) == CS5) || ((cflag & CSIZE) == CS6))
    tty->termios->c_cflag = ((cflag & ~CSIZE) |
			     (old_termios->c_cflag & CSIZE));

  configure_nr_port(info);

  /* Handle transition to B0 status */
  if ((old_termios->c_cflag & CBAUD) && !(tty->termios->c_cflag & CBAUD))
    {
      info->modem_control_register &= ~(MCR_RTS_SET_MASK | MCR_DTR_SET_MASK);
      info->update_mcr = 1;
    }

  /* Handle transition away from B0 status */
  if ((info->si->rsmode[info->portnum] != 0x85) &&
      !(old_termios->c_cflag & CBAUD) && (tty->termios->c_cflag & CBAUD))
    {
      info->modem_control_register |= MCR_RTS_SET_ON | MCR_DTR_SET_ON;
      info->update_mcr = 1;
    }

  if ((old_termios->c_cflag & CRTSCTS) && !(tty->termios->c_cflag & CRTSCTS))
    {
      tty->hw_stopped = 0;
      nrp_start(tty);
    }
}

/*
 *   TTY send break handler
 */
static void nrp_break(struct tty_struct *tty, int break_state)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  int orig;

  if (nslink_paranoia_check(info, "nrp_break"))
    return;

  orig = info->modem_control_register;
  
  if (break_state == -1)
    info->modem_control_register |= MCR_BREAK_SET_ON;
  else
    info->modem_control_register &= ~MCR_BREAK_SET_ON;
  
  if (orig != info->modem_control_register)
    {
      info->update_mcr = 1;
      info->si->need_to_send = 1;
      if (info->si->use_tcp == 0)
        si_timer((unsigned long) info->si);
    }
}

/* 
 *  Use bits 0x200, 0x400 and 0x800 as termios Comtrol flags, the are undefined in Linux
 */
#define TIOCM_TX_FLOWED_OFF  0x0200
#define TIOCM_SET_TXOFF      0x0400
#define TIOCM_CLEAR_TXOFF    0x0800
#define TIOCM_ICTO_OCCURRED  0x1000
#define TIOCM_CLEAR_ICTO     0x2000

static int get_modem_info(struct nr_port *info, unsigned int *value)
{
  unsigned int result;
  int mcr, msr;

  mcr = info->modem_control_register;
  msr = info->si->port_state[info->portnum].modemStatusRegister;

  result = ((mcr & MCR_RTS_SET_ON) ? TIOCM_RTS : 0)
    | ((mcr & MCR_DTR_SET_ON) ? TIOCM_DTR : 0)
    | ((msr & MSR_CD_ON) ? TIOCM_CAR : 0)
    | ((msr & MSR_RING_ON) ? TIOCM_RNG : 0)
    | ((msr & MSR_DSR_ON) ? TIOCM_DSR : 0)
    | ((msr & MSR_CTS_ON) ? TIOCM_CTS : 0)
    | ((msr & MSR_TX_FLOWED_OFF) ? TIOCM_TX_FLOWED_OFF : 0)
    | ((msr & MSR_ICTO_OCCURRED) ? TIOCM_ICTO_OCCURRED : 0);

  if (copy_to_user(value, &result, sizeof(int)))
    return -EFAULT;
  return 0;
}

static int set_modem_info(struct nr_port *info, unsigned int cmd, unsigned int *value)
{
  unsigned int arg;
  int mcr;

  if (copy_from_user(&arg, value, sizeof(int)))
    return -EFAULT;

  mcr = info->modem_control_register;

  switch (cmd)
    {
    case TIOCMBIS:
      if (info->si->rsmode[info->portnum] != 0x85)
	{
	  if (arg & TIOCM_RTS)
	    {
	      mcr |= MCR_RTS_SET_ON;
	    }
	  if (arg & TIOCM_DTR)
	    mcr |= MCR_DTR_SET_ON;
	}
      break;
    case TIOCMBIC:
      if (arg & TIOCM_RTS)
	mcr &= ~MCR_RTS_SET_ON;
      if (arg & TIOCM_DTR)
	mcr &= ~MCR_DTR_SET_ON;
      break;
    case TIOCMSET:
      if (info->si->rsmode[info->portnum] != 0x85)
	{
	  mcr = ((mcr & ~(MCR_RTS_SET_ON | MCR_DTR_SET_ON))
		 | ((arg & TIOCM_RTS) ? MCR_RTS_SET_ON : 0)
		 | ((arg & TIOCM_DTR) ? MCR_DTR_SET_ON : 0));

	  if (arg & TIOCM_SET_TXOFF) {
	      info->action_register = ACT_SET_TX_XOFF;
	      info->update_action = 1;
	  }

	  if (arg & TIOCM_CLEAR_TXOFF) {
	      info->action_register = ACT_CLEAR_TX_XOFF;
	      info->update_action = 1;
	  }

	  if (arg & TIOCM_CLEAR_ICTO) {
//	      printk(KERN_INFO "nslink: Clearing MSR_ICTO_OCCURRED \n"); 
	      info->si->port_state[info->portnum].modemStatusRegister &= ~MSR_ICTO_OCCURRED;
	  }
	}
      break;
    default:
      return -EINVAL;
    }

  if (mcr != info->modem_control_register)
    {
      info->modem_control_register = mcr;
      info->update_mcr = 1;
      info->si->need_to_send = 1;
      if (info->si->use_tcp == 0)
        si_timer((unsigned long) info->si);
    }

  return 0;
}

#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)

/*
 *  Returns the state of the serial modem control lines.  These next 2 functions 
 *  are the way kernel versions > 2.5 handle modem control lines rather than IOCTLs.
 */
static int nrp_tiocmget(struct tty_struct *tty, struct file *file)
{
  struct nr_port *info = (struct nr_port *)tty->driver_data;
  unsigned int result;
  int mcr, msr;

  mcr = info->modem_control_register;
  msr = info->si->port_state[info->portnum].modemStatusRegister;

  result = ((mcr & MCR_RTS_SET_ON) ? TIOCM_RTS : 0)
    | ((mcr & MCR_DTR_SET_ON) ? TIOCM_DTR : 0)
    | ((msr & MSR_CD_ON) ? TIOCM_CAR : 0)
    | ((msr & MSR_RING_ON) ? TIOCM_RNG : 0)
    | ((msr & MSR_DSR_ON) ? TIOCM_DSR : 0)
    | ((msr & MSR_CTS_ON) ? TIOCM_CTS : 0);

  return result;
}

/* 
 *  Sets the modem control lines
 */
static int nrp_tiocmset(struct tty_struct *tty, struct file *file,
		    unsigned int set, unsigned int clear)
{
  struct nr_port *info = (struct nr_port *)tty->driver_data;
  int mcr;

  mcr = info->modem_control_register;

  //  If the port is not in RS485 mode, set the modem control lines
  if (info->si->rsmode[info->portnum] != 0x85)
    {
      if (set & TIOCM_RTS)
	mcr |= MCR_RTS_SET_ON;
      if (set & TIOCM_DTR)
	mcr |= MCR_DTR_SET_ON;
      if (clear & TIOCM_RTS)
	mcr &= ~MCR_RTS_SET_ON;
      if (clear & TIOCM_DTR)
	mcr &= ~MCR_DTR_SET_ON;
    }	

  if (mcr != info->modem_control_register)
    {
      info->modem_control_register = mcr;
      info->update_mcr = 1;
      info->si->need_to_send = 1;
      if (info->si->use_tcp == 0)
        si_timer((unsigned long) info->si);
    }

  return 0;
}

#endif /*  Linux > 2.5 */


static int get_config(struct nr_port *info, struct si_config *retinfo)
{
  struct si_config tmp;

  if (!retinfo)
    return -EFAULT;
  memset(&tmp, 0, sizeof(tmp));
  tmp.line = info->line;
  tmp.flags = info->flags;
  tmp.close_delay = info->close_delay;
  tmp.closing_wait = info->closing_wait;
#if 0
  tmp.port = rcktpt_io_addr[(info->line >> 5) & 3];
#else
  tmp.port = 0;
#endif

  if (copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
    return -EFAULT;
  return 0;
}

static int set_config(struct nr_port *info, struct si_config *new_info)
{
  struct si_config new_serial;

  if (copy_from_user(&new_serial, new_info, sizeof(new_serial)))
    return -EFAULT;

#ifdef CAP_SYS_ADMIN
	if (!capable(CAP_SYS_ADMIN))
#else
	if (!suser())
#endif
    {
      if ((new_serial.flags & ~SI_USR_MASK) !=
	  (info->flags & ~SI_USR_MASK))
	return -EPERM;
      info->flags = ((info->flags & ~SI_USR_MASK) |
		     (new_serial.flags & SI_USR_MASK));
      configure_nr_port(info);
      return 0;
    }

  info->flags = ((info->flags & ~SI_FLAGS) |
		 (new_serial.flags & SI_FLAGS));
  info->close_delay = new_serial.close_delay;
  info->closing_wait = new_serial.closing_wait;

  if ((info->flags & SI_SPD_MASK) == SI_SPD_HI)
    info->tty->alt_speed = 57600;
  if ((info->flags & SI_SPD_MASK) == SI_SPD_VHI)
    info->tty->alt_speed = 115200;
  if ((info->flags & SI_SPD_MASK) == SI_SPD_SHI)
    info->tty->alt_speed = 230400;
  if ((info->flags & SI_SPD_MASK) == SI_SPD_WARP)
    info->tty->alt_speed = 460800;

  configure_nr_port(info);
  return 0;
}

static int get_serial_info(struct nr_port *info,
			   struct serial_struct * retinfo)
{
  struct serial_struct tmp;
  if (!retinfo)
    return -EFAULT;
  memset(&tmp, 0, sizeof(tmp));
  if (info->low_latency)
    tmp.flags |= ASYNC_LOW_LATENCY;
  tmp.line = info->portnum;
  tmp.irq  = info->inter_char_timeout;
  if (info->si->use_tcp)
    tmp.port = ntohl(info->si->ip_addr);
  else
    tmp.port = ((info->si->ether_addr[2] << 24)|
                (info->si->ether_addr[3] << 16)|
                (info->si->ether_addr[4] << 8)|
                (info->si->ether_addr[5] << 0));
  if (copy_to_user(retinfo,&tmp,sizeof(*retinfo)))
    return -EFAULT;
  return 0;
}

static int set_serial_info(struct nr_port *info,
			   struct serial_struct * new_info)
{
  struct serial_struct new_serial;
  if (copy_from_user(&new_serial,new_info,sizeof(new_serial)))
    return -EFAULT;
  info->low_latency = ((new_serial.flags & ASYNC_LOW_LATENCY) != 0);
  if (info->tty)
    info->tty->low_latency = info->low_latency;
  info->inter_char_timeout = new_serial.irq;
  if (info->si->arch_id == 3 || info->si->arch_id ==2)
    info->update_ic_timeout = 1;
  return 0;
}

/*
 *  Handler for IOCTL calls made through the tty device
 */
static int nrp_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;

  if (nslink_paranoia_check(info, "nrp_ioctl"))
    return -ENODEV;
  
  if (info->remote_status != REM_PORT_ASSIGNED)
    return -EIO;
  
  switch (cmd)
    {
    case TIOCGSERIAL:
      return get_serial_info(info,(struct serial_struct*)arg);
    case TIOCSSERIAL:
      return set_serial_info(info,(struct serial_struct*)arg);
    case TIOCMGET:
      return get_modem_info(info, (unsigned int *) arg);
    case TIOCMBIS:
    case TIOCMBIC:
    case TIOCMSET:
      return set_modem_info(info, cmd, (unsigned int *) arg);
    case SI_GET_CONFIG:
      return get_config(info, (struct si_config *) arg);
    case SI_SET_CONFIG:
      return set_config(info, (struct si_config *) arg);
    case SI_GET_CTL_MAJOR:
      if (copy_to_user((void *) arg, &control_major, sizeof(int)))
	return -EFAULT;
    case SI_SET_BOX_CONFIG:
      return init_hub_state((struct si_box_config *) arg, 1);
      
    case SI_GET_BOX_CONFIG:
      return get_box_config((struct si_box_config *) arg);
      
    case SI_GET_BOX_REPORT:
      return get_box_report((struct si_box_report *) arg);
      
    case SI_LOAD_CODE4_8:
    case SI_LOAD_CODE2:
    default:
      return -ENOIOCTLCMD;
    }
  return 0;
}

#if (defined(NSLINK_DEBUG_FLOW) || defined(NSLINK_DEBUG_THROTTLE))
static char *nrp_tty_name(struct tty_struct *tty, char *buf)
{
  if (tty)
    sprintf(buf, "%s%d", TTY_DRIVER_NAME(tty), 
	    MINOR(tty->device) - TTY_DRIVER_MINOR_START(tty) + TTY_DRIVER_NAME_BASE);
  else
    strcpy(buf, "NULL tty");
  return buf;
}
#endif

static void nrp_send_xchar(struct tty_struct *tty, char ch)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;

  if (nslink_paranoia_check(info, "nrp_send_xchar"))
    return;
  
  info->x_char = ch;
  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);
}

static void nrp_throttle(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
#ifdef NSLINK_DEBUG_THROTTLE
  char buf[64];

  printk(KERN_INFO "nslink: throttle %s: %d....\n", nrp_tty_name(tty, buf),
	 tty->ldisc.chars_in_buffer(tty));
#endif

  if (nslink_paranoia_check(info, "nrp_throttle"))
    return;

  if (I_IXOFF(tty))
    nrp_send_xchar(tty, STOP_CHAR(tty));
}

static void nrp_unthrottle(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
#ifdef NSLINK_DEBUG_THROTTLE
  char buf[64];

  printk(KERN_INFO "nslink: unthrottle %s: %d....\n", nrp_tty_name(tty, buf),
	 tty->ldisc.chars_in_buffer(tty));
#endif

  if (nslink_paranoia_check(info, "nrp_throttle"))
    return;
  
  if (I_IXOFF(tty))
    nrp_send_xchar(tty, START_CHAR(tty));
}

/*
 * ------------------------------------------------------------
 * nrp_stop() and nrp_start()
 *
 * This routines are called before setting or resetting tty->stopped.
 * They enable or disable transmitter interrupts, as necessary.
 * ------------------------------------------------------------
 */
static void nrp_stop(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
#ifdef NSLINK_DEBUG_FLOW
  char buf[64];

  printk(KERN_INFO "nslink: stop %s: %d....\n", nrp_tty_name(tty, buf), info->xmit_cnt);
#endif

  if (nslink_paranoia_check(info, "nrp_stop"))
    return;

  /*
   * Would be nice if there was a way to stop the transmitter
   * immediately.
   */
}

static void nrp_start(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
#ifdef NSLINK_DEBUG_FLOW
  char buf[64];

  printk(KERN_INFO "nslink: start %s: %d....\n", nrp_tty_name(tty, buf),
	 info->xmit_cnt);
#endif

  if (nslink_paranoia_check(info, "nrp_start"))
    return;

  /*
   * Start the transmitter remotely, if we ever add a stop
   * transmitter to nrp_stop.
   */

  if (info->xmit_cnt <= 0 || tty->stopped || tty->hw_stopped ||
      !info->xmit_buf)
    return;

  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);
}

/*
 *  Waits until the serial port in the remote hub had transmitted all 
 *  the data that this tty driver has sent to it.
 */
static void nrp_wait_until_sent(struct tty_struct *tty, int timeout)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  unsigned long orig_jiffies;
  int check_time, exit_time;
  int txcnt;

  if (nslink_paranoia_check(info, "nrp_wait_until_sent"))
    return;
  
  if (info->remote_status != REM_PORT_ASSIGNED)
    return;

  orig_jiffies = jiffies;
#ifdef NSLINK_DEBUG_WAIT_UNTIL_SENT
  printk(KERN_INFO "nslink: In NRP_wait_until_sent(%d) (jiff=%lu)...", timeout,
	 jiffies);
  printk(KERN_INFO "nslink: cps=%d...", info->cps);
#endif
  while (1)
    {
      txcnt = get_tx_cnt_remote(&info->si->port_state[info->portnum]);
      if (!txcnt)
	{
	  /*
	   * Ideal to check if the last character has
	   * shifted all the way out, but there's no way
	   * to do that right now.
	   */
	  break;
	}
      else
	{
          // check in 1/10 second intervals.  Used to base
          // interval on tx byte count, but if there is a
          // flush in progress, that's no good.
          check_time = HZ/10;
	}
      if (timeout)
	{
	  exit_time = orig_jiffies + timeout - jiffies;
	  if (exit_time <= 0)
	    break;
	  if (exit_time < check_time)
	    check_time = exit_time;
	}
      if (check_time == 0)
	check_time = 1;
#ifdef NSLINK_DEBUG_WAIT_UNTIL_SENT
      printk(KERN_INFO "nslink: txcnt = %d (jiff=%lu,check=%d)...", txcnt,
	     jiffies, check_time);
#endif
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout(check_time);
      if (signal_pending(current))
	break;
    }
  set_current_state(TASK_RUNNING);
#ifdef NSLINK_DEBUG_WAIT_UNTIL_SENT
  printk(KERN_INFO "nslink: txcnt = %d (jiff=%lu)...done\n", txcnt, jiffies);
#endif
}

/*
 * nrp_hangup() --- called by tty_hangup() when a hangup is signaled.
 */
static void nrp_hangup(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  unsigned long flags;

  if (nslink_paranoia_check(info, "nrp_hangup"))
    return;

#if (defined(NSLINK_DEBUG_OPEN) || defined(NSLINK_DEBUG_HANGUP))
  printk(KERN_INFO "nslink: nrp_hangup of ttySI%d...", info->line);
#endif
  /*
   * If the port is in the process of being closed, just force
   * the transmit buffer to be empty, and let nrp_close handle
   * the clean up.
   */
  if (info->flags & SI_CLOSING)
    {
      spin_lock_irqsave(&info->slock, flags);
      info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
      spin_unlock_irqrestore(&info->slock, flags);

      wake_up_interruptible(&tty->write_wait);
      return;
    }

  //  Use spinlock to protect info->count
  spin_lock_irqsave(&info->slock, flags);
  if (info->count)
    {

#if ((LINUX_VERSION_CODE < VERSION_CODE(2,3,0)) && defined(MODULE))
	MOD_DEC_USE_COUNT;
#endif

      atomic_dec(&nrp_num_ports_open);
    }

  info->count = 0;
  spin_unlock_irqrestore(&info->slock, flags);

  info->flags &= ~SI_NORMAL_ACTIVE;
  info->tty = 0;

  info->action_register = ACT_FLUSH_INPUT | ACT_FLUSH_OUTPUT;
  info->update_action = 1;
  info->si->need_to_send = 1;

  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);

  info->flags &= ~SI_INITIALIZED;

  wake_up_interruptible(&info->open_wait);
}

/*
 *  TTY write char routine.  Write buffer and counters protected by spinlocks.
 */
static void nrp_put_char(struct tty_struct *tty, unsigned char ch)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  unsigned long flags;

  if (nslink_paranoia_check(info, "nrp_put_char"))
    return;
  
  if (info->remote_status != REM_PORT_ASSIGNED)
    return;

  if (info->xmit_cnt >= XMIT_BUF_SIZE - 1)
      return;

#ifdef NSLINK_DEBUG_WRITE
  printk(KERN_INFO "nslink: nrp_put_char %c...", ch);
#endif

  down_interruptible(&info->write_sem);
  spin_lock_irqsave(&info->slock, flags);

  info->xmit_buf[info->xmit_head++] = ch;
  info->xmit_head &= SERIAL_XMIT_SIZE - 1;
  info->xmit_cnt++;

  spin_unlock_irqrestore(&info->slock, flags);
  up(&info->write_sem);
}

/*
 *  TTY flush chars routine
 */
static void nrp_flush_chars(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  struct si_port_state *pstate;
  int space = 0;

  if (nslink_paranoia_check(info, "nrp_flush_chars"))
    return;

  pstate = &info->si->port_state[info->portnum];
  if (pstate)
    space = get_tx_space_remote(pstate);

  if (space <= WAKEUP_CHARS || info->xmit_cnt <= 0 ||
      tty->stopped || tty->hw_stopped || !info->xmit_buf)
    return;

  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    {
      if (info->xmit_cnt > 64)
	si_timer((unsigned long) info->si);
      else
	si_set_timer(info->si, 0);
    }
}

/*
 *  TTY write routine, called when user (or kernel) app writes to a serial port.  Data
 *  is moved into the write buffer for the port (in the global info-> struct), then
 *  the HDLC or TCP function is called that causes the bytes to go over the ethernet
 *  to the device.
 *
 *  A per port write semaphore is used to protect from another process writing to
 *  this port at the same time.  This other process could be running on the other CPU
 *  or get control of the CPU if the copy_from_user() blocks due to a page fault (swapped out). 
 *  Spinlocks protect the info xmit members.
 *
 *  Inputs:  from_user set if write called from user space, data buf and count requested amount 
 *           to write
 *  Returns: Number of bytes actually written.
 */
static int nrp_write(struct tty_struct *tty, 
		     int from_user,             
		     const unsigned char *buf,  
		     int count)                 
{
    struct nr_port *info = (struct nr_port *) tty->driver_data;
    int c, retval = 0;
    unsigned long flags;

    if (count <= 0 || nslink_paranoia_check(info, "nrp_write"))
	return 0;

    if (info->si->state == SI_STATE_RESET)
	return -EIO;
    
    if (info->remote_status != REM_PORT_ASSIGNED)
	return -ENOTCONN;
    
#ifdef NSLINK_DEBUG_WRITE
    printk(KERN_INFO "nslink: nrp_write %d chars...", count);
#endif

    //  Lock the port semaphore
    down_interruptible(&info->write_sem);

    //  Loop while there is TTY data to be transfered into the driver
    while (1)  {
	if (info->tty == 0) 
	    goto end;

	c = MIN(count, MIN(XMIT_BUF_SIZE - info->xmit_cnt - 1,
			   XMIT_BUF_SIZE - info->xmit_head));
	if (c <= 0)
	    break;

	info->txBytes += c;

	//  Copy data from the app into the port's write buffer
	if (from_user) {
	  if (copy_from_user(info->xmit_buf + info->xmit_head, buf, c)) 
	    {
	      retval = -EFAULT;
	      goto end_intr;
	    } 
	}
	else {
          memcpy(info->xmit_buf + info->xmit_head, buf, c);
	}

	spin_lock_irqsave(&info->slock, flags);
	info->xmit_head = (info->xmit_head + c) & (XMIT_BUF_SIZE - 1);
	info->xmit_cnt += c;
	spin_unlock_irqrestore(&info->slock, flags);
	
	buf += c;
	count -= c;
	retval += c;
    }

end_intr:
    nrp_flush_chars(tty);

end:
    up(&info->write_sem);

    //  Call the routines to send data to hub
    if (info->tty->low_latency)   {
	info->si->need_to_send = 1;
	if (info->si->use_tcp == 0)
	    si_timer((unsigned long)info->si);
	else
	    tcp_send_remote_cmds(info->si);
    }
#ifdef NSLINK_DEBUG_WRITE
    printk(KERN_INFO "ret=%d\n", retval);
#endif
    return retval;
}


/*
 * Return the number of characters that can be sent.  We estimate
 * only using the in-memory transmit buffer only, and ignore the
 * potential space in the remote end.
 */
static int nrp_write_room(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  int ret;

  if (nslink_paranoia_check(info, "nrp_write_room"))
    return 0;

  ret = XMIT_BUF_SIZE - info->xmit_cnt - 1;
  if (ret < 0)
    ret = 0;
#ifdef NSLINK_DEBUG_WRITE
  printk(KERN_INFO "nslink: nrp_write_room returns %d...", ret);
#endif
  return ret;
}

/*
 * Return the number of characters in the buffer.  Again, this only
 * counts those characters in the in-memory transmit buffer.
 */
static int nrp_chars_in_buffer(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;

  if (nslink_paranoia_check(info, "nrp_chars_in_buffer"))
    return 0;

#ifdef NSLINK_DEBUG_WRITE
  printk(KERN_INFO "nslink: nrp_chars_in_buffer returns %d...", info->xmit_cnt);
#endif
  return info->xmit_cnt;
}

/*
 *  TTY flush buffer routine.  Removes all bytes from the nr_port xmit_buf.
 */
static void nrp_flush_buffer(struct tty_struct *tty)
{
  struct nr_port *info = (struct nr_port *) tty->driver_data;
  unsigned long flags;

  if (nslink_paranoia_check(info, "nrp_flush_buffer"))
    return;

  info->action_register = ACT_FLUSH_OUTPUT;
  info->update_action = 1;
  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    si_timer((unsigned long) info->si);

  spin_lock_irqsave(&info->slock, flags);
  info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
  spin_unlock_irqrestore(&info->slock, flags);

  wake_tty_write(tty);
}

/* 
 * /proc fs routines
 */
static inline int line_info(char *buf, struct si_state *si, int line)
{
  struct si_port_state *pstate;
  struct nr_port *info;
  char stat_buf[40];
  int ret;
  unsigned int control, status;

  ret = sprintf(buf, "%d: ", line + si->base_port);

  info = nrp_table[line + si->base_port];
  pstate = si->port_state + line;

  /*
   * Figure out the current RS-232 lines
   */
  status = pstate->modemStatusRegister;
  control = info ? info->modem_control_register : 0;
  stat_buf[0] = 0;
  stat_buf[1] = 0;
  if (control & MCR_DTR_SET_MASK)
    strcat(stat_buf, "|RTS");
  if (status & MSR_CTS_ON)
    strcat(stat_buf, "|CTS");
  if (control & MCR_DTR_SET_MASK)
    strcat(stat_buf, "|DTR");
  if (status & MSR_DSR_ON)
    strcat(stat_buf, "|DSR");
  if (status & MSR_CD_ON)
    strcat(stat_buf, "|CD");
  if (status & MSR_TX_FLOWED_OFF)
    strcat(stat_buf, "|TXOFF");
  if (status & MSR_RING_ON)
    strcat(stat_buf, "|RING");

  if (info && info->baud_rate)
    ret += sprintf(buf + ret, " baud:%d", info->baud_rate);

#if 0
  ret += sprintf(buf + ret, " tx:%d rx:%d",
		 state->icount.tx, state->icount.rx);

  if (state->icount.frame)
    ret += sprintf(buf + ret, " fe:%d", state->icount.frame);

  if (state->icount.parity)
    ret += sprintf(buf + ret, " pe:%d", state->icount.parity);

  if (state->icount.brk)
    ret += sprintf(buf + ret, " brk:%d", state->icount.brk);

  if (state->icount.overrun)
    ret += sprintf(buf + ret, " oe:%d", state->icount.overrun);
#endif

  /*
   * Last thing is the RS-232 status lines
   */
  ret += sprintf(buf + ret, " %s\n", stat_buf + 1);
  return ret;
}

static int nrp_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
  int i, j, len = 0, l;
  off_t begin = 0;
  struct si_state *si;

  len += sprintf(page, "serinfo:1.0 driver:%s\n", SI_VERSION);
  for (i = 0; i < MAX_NRP_SIS; i++)
    {
      si = conc_state[i];
      if (!si)
	continue;
      for (j = 0; j < si->num_ports; j++)
	{
	  l = line_info(page + len, si, j);
	  len += l;
	  if (len + begin > off + count)
	    goto done;
	  if (len + begin < off)
	    {
	      begin += len;
	      len = 0;
	    }
	}
    }
  *eof = 1;
done:
  if (off >= len + begin)
    return 0;
  *start = page + (begin - off);
  return ((count < begin + len - off) ? count : begin + len - off);
}


/********************************************************************************/
/*                                                                              */
/*        NSLinkctl device functions                                            */
/*                                                                              */
/********************************************************************************/

static int ctl_open(struct inode *inode, struct file *file)
{
#if ((LINUX_VERSION_CODE < VERSION_CODE(2,3,0)) && defined(MODULE))
    MOD_INC_USE_COUNT;
#endif
  return 0;
}

static int ctl_close(struct inode *inode, struct file *file)
{
#if ((LINUX_VERSION_CODE < VERSION_CODE(2,3,0)) && defined(MODULE))
    MOD_DEC_USE_COUNT;
#endif
  return 0;
}

static unsigned int ctl_poll(struct file *file, poll_table * wait)
{
  struct si_state *si;
  unsigned int mask;
  int i;

  poll_wait(file, &wake_daemon_wait, wait);
  mask = 0;

  for (i = 0; i < MAX_NRP_SIS; i++)
    {
      si = conc_state[i];
      if (si && si->state == SI_STATE_DEAD)
	mask |= POLLOUT | POLLWRNORM;
    }
  return mask;
}

static int get_box_config(struct si_box_config *retinfo)
{
  struct si_box_config tmp;
  struct si_state *si;

  if (copy_from_user(&tmp, retinfo, sizeof(tmp)))
    return -EFAULT;

  if ((tmp.boxid >= MAX_NRP_SIS) || !((si = conc_state[tmp.boxid])))
    return -EINVAL;

  tmp.num_ports = si->num_ports;
  tmp.state = si->state;
  memcpy(tmp.ether_addr, si->ether_addr, 6);
  strncpy(tmp.interface, si->dev->name, sizeof(tmp.interface));

  if (copy_to_user(retinfo, &tmp, sizeof(tmp)))
    return -EFAULT;
  return 0;
}

static int get_box_report(struct si_box_report *retinfo)
{
  struct si_box_report tmp;
  int unit;

  if (copy_from_user(&tmp, retinfo, sizeof(tmp)))
    return -EFAULT;

  unit = tmp.flags;		/* kludge, but easist way to handle it */

  if ((unit >= num_box_rpts) || (unit >= MAX_NRP_SIS))
    return -EINVAL;

  tmp = si_box_rpts[unit];

  if (copy_to_user(retinfo, &tmp, sizeof(tmp)))
    return -EFAULT;
  return 0;
}

static int get_problem_boxes(int *ret_status)
{
  struct si_state *si;
  int i, status = 0;

  for (i = 0; i < MAX_NRP_SIS; i++)
    {
      si = conc_state[i];
      if (si && si->state == SI_STATE_DEAD)
	status |= 1 << i;
    }
  if (copy_to_user(ret_status, &status, sizeof(status)))
    return -EFAULT;
  return 0;
}

static int force_port_release(int line)
{
  struct nr_port *info;
  if (line < 0 || line > (MAX_NRP_SIS * MAX_NRP_PORTS))
    return -EINVAL;
  info = nrp_table[line];
  if (!info || !info->si)
    return -EINVAL;
  printk(KERN_INFO "nslink: Forcing release of /dev/ttySI%d\n",line);
  info->send_force_release = 1;
  info->force_release_acked = 0;
  info->si->need_to_send = 1;
  if (info->si->use_tcp == 0)
    mod_timer(&info->si->timer, jiffies + HZ/50);  //  Set the timer to 20ms from now
  interruptible_sleep_on_timeout(&info->release_wait,3*HZ);
  if (info->force_release_acked)
    return 0;
  else
    return -ETIME;
}

static int debug_misc(void)
{
#if 0
  struct si_state *si;

  si = conc_state[0];
  printk(KERN_INFO "nslink: si->rtq.qlen = %d\n", si->rtq.qlen);
#endif
  return 0;
}

/*
 *  Sets number of open ports to zero so the can_unload() returns true, allowing the module
 *  to unload.  Kind of a hack, we have continual problems with this use count stuff.  Be careful
 *  not to do this when ports really are open.
 */
static int release_module(void)
{
    atomic_set(&nrp_num_ports_open, 0);
    return 0;
}

/*
 *  Handler for IOCTL calls made through the NSLinkctl device
 */
static int ctl_ioctl(struct inode *inode, 
		     struct file *file, 
		     unsigned int cmd, 
		     unsigned long arg)
{
  switch (cmd)
    {
    case SI_RELEASE_MODULE:
       return release_module();
      
    case SI_SET_BOX_CONFIG:
       return init_hub_state((struct si_box_config *) arg, 1);
      
    case SI_GET_BOX_CONFIG:
      return get_box_config((struct si_box_config *) arg);
      
    case SI_GET_BOX_REPORT:
      return get_box_report((struct si_box_report *) arg);
      
    case SI_GET_PROBLEM_BOXES:
      return get_problem_boxes((int *) arg);
      
    case SI_DEBUG_MISC:
      return debug_misc();
      
    case SI_SET_TCP:
      return set_tcp((struct tcp_mode_params *) arg);
      break;
      
    case SI_TCP_TXQUEUE_WAIT:
      return tcp_txqueue_wait((tTcpTxQueueWait*) arg);
      break;
      
    case SI_SET_MAC:
      return set_mac((struct mac_mode_params *) arg);
      break;
      
    case SI_FORCE_PORT_RELEASE:
      return force_port_release(arg);
      break;
      
    case SI_LOAD_CODE4_8:
    case SI_LOAD_CODE2:
    default:
      return -EINVAL;
    }
  return 0;
}

/*
 *  Sets the MAC mode parameters for the driver.  Called from the NSLinkctl device IOCTL.
 */
static int set_mac(struct mac_mode_params *in)
{
  struct si_state *si;
  struct mac_mode_params mac_params;

  if (copy_from_user(&mac_params, (void *) in, sizeof(struct mac_mode_params)))
    return -EFAULT;
  
  if (mac_params.boxid >= MAX_NRP_SIS)
    return -EINVAL;

  si = conc_state[mac_params.boxid];
  
  if (!si)
    return -EFAULT;
  
  si->state = SI_STATE_BOOTED;
  si_set_timer(si, HZ / 50);         //  20ms wakeup call
  
  //  Sleep until woken.  The daemon configures the box, then sends the set_mac
  //  ioct.  At this point, the hub begins functioning and the daemon sleeps (here).
  //  It is woken (in si_timer) only if the state goes to SI_STATE_RESET, at which 
  //  point it it the daemon's job to reboot the hub and restart it.
  interruptible_sleep_on(&si->waitqueue);
  return 0;
}

/*
 *  File operations struct for the NSLinkctl device
 */
static struct file_operations ctl_fops = 
{
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
    owner: THIS_MODULE,
#endif     
    poll: ctl_poll,
    ioctl: ctl_ioctl,
    open:  ctl_open,
    release: ctl_close,
};

/********************************************************************************/
/*                                                                              */
/*        TCP FUNCTIONS                                                         */
/*                                                                              */
/********************************************************************************/

/*
 *  Stores a fragment in the fragment buffer.  A fragment is when a partial 
 *  admin/async command is received, which happens in tcp mode.  For example,
 *  a RK_PORT_SET command is the command which should have 1 byte of associated
 *  data (total length 2).  If the data length is 1, then this is a fragment, the
 *  next byte must be in the next tcp packet.  This function stores the fragment
 *  and set the tcp_frag_len.  This variable acts as a flag indicating that there
 *  is a stored fragment.
 */
static void tcp_store_fragment(struct si_state *si, unsigned char *buf, int len)
{
  memcpy(si->tcp_k_rfrag, buf, len);
  si->tcp_frag_len = len;
  return;
}

static void tcp_process_packet(int len, struct si_state *si)
{

  unsigned char *buf;

  //printk(KERN_INFO "nslink: tcp_process_packet(%d,%p)\n",len,si);

  if (si->tcp_frag_len)
    {
      buf = si->tcp_k_rfrag;
      len += si->tcp_frag_len;
      si->tcp_frag_len = 0;
    }
  else
    buf = si->tcp_k_rbuf;

  si_parse_input_pkt(si, buf, len);

  return;
}

static void tcp_check_rx(struct si_state *si)
{
  struct iovec iov;
  struct msghdr msg;
  int rc, amount, pktlen;

  rc = si->SOCK_PROT(tcp_sk)->ioctl(si->tcp_sk, TIOCINQ, (unsigned long) si->tcp_u_rbuf);

  if (rc)
    {
      printk(KERN_INFO "nslink: TIOCINQ err %d\n", rc);
      return;
    }
  if (copy_from_user(&amount, si->tcp_u_rbuf, sizeof(int)))
    {
      printk(KERN_INFO "nslink: CheckRxTcp copy from user failed\n");
      return;
    }

  while (amount)
    {
      if (amount > si->tcp_rbuf_size)
        pktlen = si->tcp_rbuf_size;
      else
        pktlen = amount;
      
      iov.iov_base = si->tcp_u_rbuf;
      iov.iov_len = pktlen;
      msg.msg_name = NULL;
      msg.msg_iov = &iov;
      msg.msg_iovlen = 1;
      msg.msg_control = NULL;

#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)
      rc = si->SOCK_PROT(tcp_sk)->recvmsg(NULL, si->tcp_sk, &msg, pktlen, 1, 0, NULL);
#else
      rc = si->SOCK_PROT(tcp_sk)->recvmsg(si->tcp_sk, &msg, pktlen, 1, 0, NULL);
#endif

      if (rc < 0)
	{
	  printk(KERN_INFO "nslink: CheckRxTcp recvmsg err %d\n", rc);
	  return;
	}
      if (rc != pktlen)
        printk(KERN_INFO "nslink: CheckRxTcp rc %d, pktlen %d\n", rc, pktlen);

      /*
       *  If tcp_frag_len is non-zero, then there is a stored fragment from the last
       *  packet.  Add the incoming data to this fragment and process that.  If no fragment,
       *  process the incoming data.
       */
      if (si->tcp_frag_len)
	  rc = copy_from_user(&si->tcp_k_rfrag[si->tcp_frag_len], si->tcp_u_rbuf, pktlen);
      else
	  rc = copy_from_user(si->tcp_k_rbuf, si->tcp_u_rbuf, pktlen);
      
      if (rc)
	{
	  printk(KERN_INFO "nslink: CheckRxTcp copy from user failed\n");
	  return;
	}

      si->tcp_to_rcvp += 1;
      tcp_process_packet(pktlen, si);
      amount -= pktlen;
    }

  return;
}

/*
 *  Send data using a kernel mode TCP socket.  Data to send is int si->tcp_k_wbuf.
 */
static void tcp_send_packet(struct si_state *si, int len)
{
  int rc;
  struct iovec iov;
  struct msghdr msg;
  mm_segment_t oldfs;

  //printk(KERN_INFO "nslink: tcp_send_packet(%p, %d)\n",si,len);

  if (!len)
    return;
  
  iov.iov_base = si->tcp_k_wbuf;
  iov.iov_len = len;
  msg.msg_name = NULL;
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_flags = 0;

  oldfs = get_fs(); 
  set_fs(KERNEL_DS);

#if LINUX_VERSION_CODE > VERSION_CODE(2,5,0)
  rc = si->SOCK_PROT(tcp_sk)->sendmsg(NULL, si->tcp_sk, &msg, len);
#else
  rc = si->SOCK_PROT(tcp_sk)->sendmsg(si->tcp_sk, &msg, len);
#endif

  set_fs(oldfs);

  if (rc == len)
    si->tcp_to_xmtp += 1;
  else
    printk(KERN_INFO "nslink: TcpSendPacket sendmsg err, %d\n", rc);

  return;
}

static int tcp_room(struct si_state *si, int bi, int cnt)
{
  if ((bi + cnt) > si->tcp_wbuf_size)
    {
      tcp_send_packet(si, bi);
      return 1;
    }
  return 0;
}

static int tcp_out_data(unsigned char *buf, int bi, struct nr_port *info,
		    struct si_port_state *pstate, struct si_state *si)
{
  int lenbi, xmitcnt, space, totlen;
  struct tty_struct *tty;
  unsigned long flags;

  buf[bi++] = RK_DATA_BLK;
  lenbi = bi;
  bi += 2;

  totlen = 0;
  if (info->x_char)
    {
      buf[bi++] = info->x_char;
      totlen += 1;
      pstate->nPutRemote += 1;
      info->x_char = 0;
    }

  //  Copy the data into buf, spinlock to protect xmit_tail and xmit_cnt
  spin_lock_irqsave(&info->slock, flags);
  do
    {
      xmitcnt = si->tcp_wbuf_size - bi;
      if (xmitcnt > info->xmit_cnt)
        xmitcnt = info->xmit_cnt;
      
      if (xmitcnt > (XMIT_BUF_SIZE - info->xmit_tail))
        xmitcnt = XMIT_BUF_SIZE - info->xmit_tail;

      space = get_tx_space_remote(pstate);
      
      if (xmitcnt > space)
        xmitcnt = space;

      if (xmitcnt)
	{
	  memcpy(&buf[bi], (info->xmit_buf + info->xmit_tail), xmitcnt);
	  bi += xmitcnt;
	  info->xmit_tail += xmitcnt;
	  info->xmit_tail &= (XMIT_BUF_SIZE - 1);
	  info->xmit_cnt -= xmitcnt;
	  totlen += xmitcnt;
	  pstate->nPutRemote += (__u16) xmitcnt;
	}
    }
  while (xmitcnt);
  spin_unlock_irqrestore(&info->slock, flags);

  if (totlen)
    {
      buf[lenbi++] = totlen & 0xff;
      buf[lenbi++] = totlen >> 8;
      totlen += 3;
    }

  tty = info->tty;
  if ((info->xmit_cnt < WAKEUP_CHARS) && tty)
    wake_tty_write(tty);

  return totlen;
}

/*
 *  Depending on the si-> tcp flags, puts cmds into the tcp write buffer,
 *  then sends the tcp packet.
 */
static void tcp_send_remote_cmds(struct si_state *si)
{
  int pi, bi, portset, dc;
  unsigned char *buf;
  struct si_port_state *pstate;
  struct nr_port *info;

  // todo: add mutex?

  //printk(KERN_INFO "nslink: tcp_send_remote_cmds(%p)\n",si);
  
  if (!sock_wspace(si->tcp_sk))
    {
      printk(KERN_INFO "nslink: TcpLoop OUTQ empty\n");
      return;
    }
  
  buf = si->tcp_k_wbuf;
  si->need_to_send = 0;
  bi = 0;
  
  if (si->need_to_send_connect_check)
    {
      buf[bi++] = RK_CONNECT_CHECK;
      si->need_to_send_connect_check = 0;
#ifdef NSLINK_DEBUG_KEEPALIVE
      printk(KERN_INFO "nslink: tx RK_CONNECT_CHECK: %d\n",si->conc_id);
#endif
    }
  
  if (si->update_timeout)
    {
      if (tcp_room(si, bi, 7))
        bi = 0;
      buf[bi++] = RK_TIMEOUT_SET;
      // whoever decided to put stuff on the wire in */
      // something other than network order ought to be
      // hunted down and beaten bloody.
      buf[bi++] = si->link_timeout;
      buf[bi++] = si->link_timeout >> 8;
      buf[bi++] = si->link_timeout >> 16;
      buf[bi++] = si->link_timeout >> 24;
      if ( si->scan_period && (si->arch_id == 3 || si->arch_id ==2))
        {
          buf[bi++] = RK_SCANPERIOD_SET;
          buf[bi++] = si->scan_period;
        }
      si->update_timeout = 0;
    }
 
  pi = 0;
  while (pi < si->num_ports)
    {
      portset = 0;
      buf[bi++] = RK_PORT_SET;
      buf[bi++] = pi;
      
      info = nrp_table[pi + si->base_port];
      if (!info)
        continue;
      
      if (info->remote_status == REM_PORT_ASSIGNED)
	{
	  pstate = si->port_state + pi;
	  if (pstate->update_queue)
	    {
	      if (tcp_room(si, bi, 3))
		{
		  bi = 0;
		  continue;
		}
	      portset = 1;
	      buf[bi++] = RK_QIN_STATUS;
	      *(unsigned short *) &buf[bi] = pstate->nGetLocal;
	      bi += 2;
	      pstate->update_queue = 0;
	    }
	  if (info->rsmode != si->rsmode[pi])
	    {
	      if (tcp_room(si, bi, 2))
		{
		  bi = 0;
		  continue;
		}

 	      if (si->rsmode[pi] != SI_RSMODE_232) {
		  if (UnitHasNo485(si->modelID)) {
		      printk(KERN_INFO "nslink: Error - attempting to change interface mode for port ttySI%d on a model that only has RS232 ports \n", info->line);
		      si->rsmode[pi] = SI_RSMODE_232;
		  }
	      }

	      info->rsmode = si->rsmode[pi];
	      portset = 1;
	      buf[bi++] = RK_RSMODE_SET;
	      buf[bi++] = info->rsmode;
	      if (info->rsmode == SI_RSMODE_485)
		{
		  /* RS485 turn on RTS toggle */
		  info->update_control_settings = 1;
		  info->control_settings &= ~SC_FLOW_RTS_MASK;
		  info->control_settings |= SC_FLOW_RTS_ARS485;
		}
	    }
	  if (info->update_baud_rate)
	    {
	      if (tcp_room(si, bi, 5))
		{
		  bi = 0;
		  continue;
		}
	      portset = 1;
	      buf[bi++] = RK_BAUD_SET;
	      *(unsigned long *) &buf[bi] = info->baud_rate;
	      bi += 4;
	      info->update_baud_rate = 0;
	    }
	  if (info->update_special_charset)
	    {
	      if (tcp_room(si, bi, 7))
		{
		  bi = 0;
		  continue;
		}
	      portset = 1;
	      buf[bi++] = RK_SPECIAL_CHAR_SET;
              buf[bi++] = info->rx_xon;
              buf[bi++] = info->rx_xoff;
              buf[bi++] = info->tx_xon;
              buf[bi++] = info->tx_xoff;
              buf[bi++] = 0;
              buf[bi++] = 0;
	      info->update_special_charset = 0;
	    }
	  if (info->update_control_settings)
	    {
	      if (tcp_room(si, bi, 3))
		{
		  bi = 0;
		  continue;
		}
	      portset = 1;
	      buf[bi++] = RK_CONTROL_SET;
	      *(unsigned short *) &buf[bi] = info->control_settings;
	      bi += 2;
	      info->update_control_settings = 0;
	    }
	  if (info->update_mcr)
	    {
	      if (tcp_room(si, bi, 3))
		{
		  bi = 0;
		  continue;
		}
	      portset = 1;
	      buf[bi++] = RK_MCR_SET;
	      *(unsigned short *) &buf[bi] = info->modem_control_register;
	      bi += 2;
	      info->update_mcr = 0;
	    }
          if (info->update_ic_timeout)
            {
              if (tcp_room(si, bi, 5))
                {
                  bi = 0;
                  continue;
                }
              portset = 1;
              buf[bi++] = RK_IC_TIMEOUT_SET;
              *(unsigned*)&buf[bi] = info->inter_char_timeout;
              bi += 4;
              info->update_ic_timeout = 0;
            }
          // make sure update action is last of the setup commands
          // since we use action response to synchronize open().
	  if (info->update_action)
	    {
	      if (tcp_room(si, bi, 3))
		{
		  bi = 0;
		  continue;
		}
	      portset = 1;
	      buf[bi++] = RK_ACTION_SET;
	      *(unsigned short *) &buf[bi] = info->action_register;
	      bi += 2;
	      info->update_action = 0;
	      info->action_register = 0;
	    }
	  if (info->x_char ||
	      (info->xmit_cnt &&
	       !info->tty->hw_stopped && !info->tty->stopped))
	    {
	      if (tcp_room(si, bi, 4))
		{
		  bi = 0;
		  continue;
		}
	      dc = tcp_out_data(buf, bi, info, pstate, si);
	      if (dc)
		{
		  bi += dc;
		  portset = 1;
		}
	    }
	}
      else if (info->remote_status == REM_OPEN_PORT)
        {
          if (tcp_room(si, bi, 2))
            {
              bi = 0;
              continue;
            }
          portset = 1;
          buf[bi++] = RK_REMOTE_STATUS;
          buf[bi++] = REM_OPEN_PORT;
          info->remote_status = REM_WAITING_OPEN;
        }
      else if (info->remote_status == REM_RELEASE_PORT)
        {
          if (tcp_room(si, bi, 2))
            {
              bi = 0;
              continue;
            }
          portset = 1;
          buf[bi++] = RK_REMOTE_STATUS;
          buf[bi++] = REM_RELEASE_PORT;
          info->remote_status = REM_WAITING_CLOSE;
        }

      if (info->send_force_release)
        {
          if (tcp_room(si,bi,1))
            {
              bi = 0;
              continue;
            }
          portset = 1;
          buf[bi++] = RK_RELEASE_REQ;
          info->force_release_acked = 0;
          info->send_force_release = 0;
        }
      
      if (portset)
	{
	  if (bi == si->tcp_wbuf_size)
	    {
	      tcp_send_packet(si, bi);
	      bi = 0;
	      continue;
	    }
	}
      else
        bi -= 2;
      
      ++pi;
    }

  if (bi)
    tcp_send_packet(si, bi);

  return;
}

// Hack Alert!
//
// Since the TCP admin stuff expects a datagram service, we
// need a way to make sure write()s don't get contatenated some
// where in the network.  This is an ioctl() that can be use
// to wait until a TCP tx queue is empty (all data has been
// ACK'ed).

static int tcp_txqueue_wait(tTcpTxQueueWait *in)
{
  tTcpTxQueueWait req;
  struct file *file;
  struct socket *socket;
  struct sock *sock;
  struct tcp_opt *tp;
  int timeout;
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0) 
  struct inode *inode;
#endif

  if (copy_from_user(&req, (void *)in, sizeof req))
    return -EFAULT;

  file = fget(req.sock);
  
  if (!file)
    return EFAULT;
  
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0) 
  inode = file->f_dentry->d_inode;
  if (!inode || !inode->i_sock)
    {
      fput(file);
      return -EFAULT;
    }
  
  socket = &inode->u.socket_i;
  if (socket->file != file)
    {
      fput(file);
      return -EFAULT;
    }
#else
  socket = sockfd_lookup(req.sock, &timeout);
  if (timeout)
    {
      fput(file);
      return -EFAULT;
    }
#endif // if kernel < 2.5
  
  sock = (struct sock*)socket->sk;
  if (!sock)
    {
      fput(file);
      return -EFAULT;
    }

  if (SOCK_STATE(sock) == TCP_TIME_WAIT)
    return 0;

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0) 
  tp = &sock->tp_pinfo.af_tcp;
#else
  tp = tcp_sk(sock);
#endif // if kernel < 2.5

  timeout = req.msTimeout / (1000/HZ);
  
  while ((tp->write_seq - tp->snd_una) && (timeout > 0))
    {
      set_current_state(TASK_INTERRUPTIBLE);
      timeout -= (10 - schedule_timeout(10));
    }

  set_current_state(TASK_RUNNING);
  fput(file);
  if (tp->write_seq - tp->snd_una)  
    return -ETIME;
  return 0;
}


#ifndef LINUX_22
static void tcp_data_ready(struct sock *sk, int count)
{
  struct si_state *si = (struct si_state *)SOCK_DATA(sk);

  if (si && si->taskPtr)
    wake_up_process(si->taskPtr);
}
#endif

#define TRACE() printk(KERN_INFO "nslink: %s line %d\n",__FILE__,__LINE__)

/*
 *  Sets the TCP parameters for the driver for one hub.  Then goes into a polling
 *  loop, reading/writing TCP packets.  Note that the task is interruptible when
 *  schedule() is called, and the TCP stack network callback can signal this function
 *  when a TCP packet for us is received (using tcp_data_ready()).
 *
 *  Called from the NSLinkctl device IOCTL.  The nslinkd daemon starts a process
 *  for each hub, which ends up calling this function as a TCP polling loop.
 */
static int set_tcp(struct tcp_mode_params *in)
{
  struct si_state *si;
  struct tcp_mode_params tcp_params;
  struct file *file = NULL;
  struct nr_port *info;
  int i;
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0) 
  struct inode *inode;
#endif

  if (copy_from_user(&tcp_params, (void *) in, sizeof(struct tcp_mode_params)))
    return -EFAULT;
  
  if (tcp_params.boxid >= MAX_NRP_SIS)
    return -EINVAL;

  file = fget(tcp_params.sk);
  if (!file)
    return -EFAULT;

  si = conc_state[tcp_params.boxid];
  if (!si)
    {
      fput(file);
      return -EFAULT;
    }

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0) 
  inode = file->f_dentry->d_inode;
  if (!inode || !inode->i_sock)
    {
      fput(file);
      return -EFAULT;
    }

  si->tcp_sock = &inode->u.socket_i;
  if (si->tcp_sock->file != file)
    {
      fput(file);
      return -EFAULT;
    }
#else 
  si->tcp_sock = sockfd_lookup(tcp_params.sk, &i);
  if (i)
    {
      fput(file);
      return -EFAULT;
    }
#endif // if kernel < 2.5

  si->tcp_sk = si->tcp_sock->sk;
  if (!si->tcp_sk)
    {
      fput(file);
      return -EFAULT;
    }

  si->state = SI_STATE_CONNECTING;
  si->tcp_rbuf_size = tcp_params.rbuf_size;
  si->tcp_u_rbuf = tcp_params.rbuf;

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0) 
  si->saddr = si->tcp_sk->saddr;
  si->sport = si->tcp_sk->num;
  si->daddr = si->tcp_sk->daddr;
#else
  si->saddr = inet_sk(si->tcp_sk)->rcv_saddr;
  si->sport = inet_sk(si->tcp_sk)->num;
  si->daddr = inet_sk(si->tcp_sk)->daddr;
#endif // if kernel < 2.5

  si->tcp_k_rbuf = kmalloc(si->tcp_rbuf_size, GFP_KERNEL);
  if (!si->tcp_k_rbuf)
    si->tcp_sk = NULL;

  si->tcp_wbuf_size = 1400;
  si->tcp_k_wbuf = kmalloc(1400, GFP_KERNEL);

  if (!si->tcp_k_wbuf)
    si->tcp_sk = NULL;

  si->tcp_k_rfrag = kmalloc((si->tcp_rbuf_size + 2003), GFP_KERNEL);

  if (!si->tcp_k_rfrag)
    si->tcp_sk = NULL;

  si->tcp_last_port = -1;
  si->taskPtr = current;
  
#ifndef LINUX_22
  si->SOCK_DATA(tcp_sk) = si;
  si->SOCK_DATA_RDY(tcp_sk) = tcp_data_ready;
#endif  

  /* TCP Loop */
  while (si->tcp_sk)
    {
      set_current_state(TASK_INTERRUPTIBLE);
      schedule_timeout(1);

      if (signal_pending(current))
	{
          printk(KERN_INFO "nslink: TCP connection to hub %d terminated by signal\n",si->conc_id);
	  break;
	}

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0) 
      if ((si->tcp_sk == NULL) ||
	  (si->tcp_sock->state != SS_CONNECTED) ||
	  (si->SOCK_STATE(tcp_sk) != TCP_ESTABLISHED) ||
	  (si->SOCK_SHUTDOWN(tcp_sk) & SEND_SHUTDOWN) ||
	  (si->tcp_sk->saddr != si->saddr) ||
	  (si->tcp_sk->num != si->sport) || (si->tcp_sk->daddr != si->daddr))
	{
          printk(KERN_INFO "nslink: TCP connection to hub %d terminated by %s\n",si->conc_id,
                 si->tcp_sk == NULL ? "socket==NULL" :
                 si->tcp_sock->state != SS_CONNECTED ? "ss_connected==0" :
		 si->SOCK_SHUTDOWN(tcp_sk) & SEND_SHUTDOWN ? "send_shutdown==1" :
                 si->tcp_sk->saddr != si->saddr ? "bad saddr" :
                 si->tcp_sk->daddr != si->daddr ? "bad daddr" :
                 si->tcp_sk->num != si->sport ? "bad num" :
                 "?"
                 );
          si->tcp_sk = NULL;
          break;
	}
#else
      if ((si->tcp_sk == NULL) ||
	  (si->tcp_sock->state != SS_CONNECTED) ||
	  (si->SOCK_STATE(tcp_sk) != TCP_ESTABLISHED) ||
	  (si->SOCK_SHUTDOWN(tcp_sk) & SEND_SHUTDOWN) ||
	  (inet_sk(si->tcp_sk)->rcv_saddr != si->saddr) ||
	  (inet_sk(si->tcp_sk)->num != si->sport) || 
	  (inet_sk(si->tcp_sk)->daddr != si->daddr))
	{
          printk(KERN_INFO "nslink: TCP connection to hub %d terminated by %s\n",si->conc_id,
                 si->tcp_sk == NULL ? "socket==NULL" :
                 si->tcp_sock->state != SS_CONNECTED ? "ss_connected==0" :
                 si->SOCK_SHUTDOWN(tcp_sk) & SEND_SHUTDOWN ? "send_shutdown==1" :
                 inet_sk(si->tcp_sk)->rcv_saddr != si->saddr ? "bad saddr" :
		 inet_sk(si->tcp_sk)->daddr != si->daddr ? "bad daddr" :
                 inet_sk(si->tcp_sk)->num != si->sport ? "bad num" :
                 "?"
                 );
          si->tcp_sk = NULL;
          break;
	}
#endif // if kernel < 2.5
      
      tcp_check_rx(si);

      switch (si->state)
        {
         case SI_STATE_RESET:
          // box timed out.  return and let daemon try to start it
          si->tcp_sk = NULL;
          break;
              
         case SI_STATE_CONNECTING:
          /* Send assign request */
          //printk(KERN_INFO "nslink: sending ASSIGN req\n");
          si->tcp_k_wbuf[0] = RK_ASSIGN_REQ;
          tcp_send_packet(si, 1);
          si->tcp_timer = jiffies + (5 * HZ);
          si->state = SI_STATE_CONNECTING_WAIT;
          break;
          
         case SI_STATE_CONNECTING_WAIT:
          if (jiffies > si->tcp_timer)
            {
              printk(KERN_INFO "nslink: hub timed out\n");
              si->state = SI_STATE_RESET;
            }
          break;
          
         case SI_STATE_CONNECTED:
          if (si->linkTimeoutJiffies)
            {
              if (si->keep_alive_tx_timer < jiffies)
                {
                  si->need_to_send = 1;
                  si->need_to_send_connect_check = 1;
                  si->keep_alive_tx_timer += si->linkTimeoutJiffies/2;
                }
              if (si->keep_alive_timeout < jiffies)
                {
                  printk(KERN_INFO "nslink: tcp keepalive timeout on hub %d\n",si->conc_id);
                  si->state = SI_STATE_RESET;
                  si->tcp_sk = NULL;
                  set_otherclose_all_tty(si);
                  wake_all_tty(si);
                }
            }
          
          if (si->tcp_sk)
            {
              for (i=0; i < si->num_ports; ++i)
                {
                  info = nrp_table[i + si->base_port];
                  if (info && (info->magic == RPORT_MAGIC))
                    send_recv_buf_to_ldisc(info);
                }
              if (si->need_to_send)
                tcp_send_remote_cmds(si);
            }
          break;
	}
    }

  if (si->tcp_k_rbuf)
    {
      kfree(si->tcp_k_rbuf);
      si->tcp_k_rbuf = NULL;
    }

  if (si->tcp_k_wbuf)
    {
      kfree(si->tcp_k_wbuf);
      si->tcp_k_wbuf = NULL;
    }

  if (si->tcp_k_rfrag)
    {
      kfree(si->tcp_k_rfrag);
      si->tcp_k_rfrag = NULL;
    }

  set_current_state(TASK_RUNNING);

  si->tcp_u_rbuf = NULL;
  si->saddr = 0;
  si->sport = 0;
  si->daddr = 0;
  si->tcp_sk = NULL;
  si->tcp_sock = NULL;
  si->state = SI_STATE_RESET;
  si->taskPtr = NULL;

  fput(file);

  return 0;
}

static int free_si_box(int boxid)
{
  struct si_state *si;
  struct nr_port *info;
  int i;

  if (boxid >= MAX_NRP_SIS || !conc_state[boxid])
    return -EINVAL;

  si = conc_state[boxid];

  for (i = 0; i < si->num_ports; i++)
    {
      info = nrp_table[i + si->base_port];
      if (!info)
	continue;
      if (info->count || info->blocked_open)
	return -EBUSY;
    }

  if (timer_pending(&si->timer))
    del_timer(&si->timer);

  conc_state[boxid] = 0;

  for (i = 0; i < si->num_ports; i++)
    {
      info = nrp_table[i + si->base_port];
      if (info)
	{
	  nrp_table[i + si->base_port] = 0;
	  kfree(info);
	}
    }

  if (si->port_state)
    kfree(si->port_state);
  kfree(si);
  return 0;
}

/********************************************************************************/
/*                                                                              */
/*        UTILITY FUNCTIONS                                                     */
/*                                                                              */
/********************************************************************************/
static char *macToStr(unsigned char *mac)
{
  static char b[20];
  sprintf(b, "%02x:%02x:%02x:%02x:%02x:%02x",
                     mac[0],
                     mac[1],
                     mac[2],
                     mac[3],
                     mac[4],
                     mac[5]);
  return b;
}

static char *siStateStr(int i)
{
  switch (i)
    {
     case SI_STATE_CONNECTING:
     case SI_STATE_CONNECTING_WAIT:
      return "Connecting";
     case SI_STATE_CONNECTED:
      return "Connected";
     case SI_STATE_DEAD:
      return "Dead";
     case SI_STATE_RESET:
      return "Reset";
     case SI_STATE_BOOTING:
      return "Booting";
     case SI_STATE_BOOTED:
      return "Booted";
    }
  return "????";
}

static char *ipToStr(unsigned long ip)
{
  static char b[20];
  unsigned long addr = htonl(ip);
    
  sprintf(b,"%ld.%ld.%ld.%ld",
          (addr>>24)&0xff,
          (addr>>16)&0xff,
          (addr>> 8)&0xff,
          (addr>> 0)&0xff);
  return b;
}

// Tell us how many characters are pending in the remote that
// haven't yet been acknowledged as being transmited.
static inline int get_tx_cnt_remote(struct si_port_state *pstate)
{
  signed short int s = (pstate->nPutRemote - pstate->nGetRemote);
  if (s<0 || s>2000)
    {
      printk(KERN_INFO "nslink: get>put error - get=%04x put=%04x cnt=%d\n",
             pstate->nGetRemote,pstate->nPutRemote,s);
      pstate->nPutRemote = pstate->nGetRemote;
      s = 0;
    }
  return s;
}

// Return how much space is left in the remote's buffer.  We
// assume the remote has space for 1999 characters
static inline int get_tx_space_remote(struct si_port_state *pstate)
{
  int c = 1999 - get_tx_cnt_remote(pstate);
  if (c < 0)
    c = 0;
  return c;
}

static void inline wake_tty_write(struct tty_struct *tty)
{
  wake_up_interruptible(&tty->write_wait);
  if ((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc.write_wakeup)
    (tty->ldisc.write_wakeup) (tty);
}

static void inline wake_tty(struct tty_struct *tty)
{
  //printk(KERN_INFO "nslink: wake_tty(%p)\n",tty);
  wake_up_interruptible(&tty->read_wait);
  wake_tty_write(tty);
#if HavePollWait  
  wake_up_interruptible(&tty->poll_wait);
#endif		   
}

// Wake up all tty's for a hub
static void wake_all_tty(struct si_state *si)
{
  struct nr_port *info;
  int i;

  for (i = 0; i < si->num_ports; i++)
    {
      info = nrp_table[i + si->base_port];
      if (info && info->tty)
        wake_tty(info->tty);
    }
}

// Set the TTY_OTHER_CLOSED bit for all tty's for a hub
static void set_otherclose_all_tty(struct si_state *si)
{
  struct nr_port *info;
  int i;

  for (i = 0; i < si->num_ports; i++)
    {
      info = nrp_table[i + si->base_port];
      if (info && info->tty)
        set_bit(TTY_OTHER_CLOSED, &info->tty->flags);
    }
}

// Clear the TTY_OTHER_CLOSED bit for all tty's for a hub
static void clr_otherclose_all_tty(struct si_state *si)
{
  struct nr_port *info;
  int i;

  for (i = 0; i < si->num_ports; i++)
    {
      info = nrp_table[i + si->base_port];
      if (info && info->tty)
        clear_bit(TTY_OTHER_CLOSED, &info->tty->flags);
    }
}


/********************************************************************************/
/*                                                                              */
/*        /proc FUNCTIONS                                                       */
/*                                                                              */
/********************************************************************************/

#undef TraceHdlc
#undef TraceHdlcStop
#undef TraceQueue

#if NSLINK_DEBUG_TRACE_QUEUE
#define NumQueueTrace 4096 /* must be power of 2 */
#define NumQueueMask (NumQueueTrace-1)
struct queueTraceStruct
{
  unsigned long timestamp;
  int port;
  unsigned char event;
  unsigned data;
  struct si_port_state pstate;
  
}queueTraceBuffer[NumQueueTrace];
unsigned queueTraceIndex;

#define TraceQueue(info,ev,d) \
  do {\
    queueTraceBuffer[queueTraceIndex].timestamp = jiffies; \
    queueTraceBuffer[queueTraceIndex].port = info->line; \
    queueTraceBuffer[queueTraceIndex].event = ev; \
    queueTraceBuffer[queueTraceIndex].data = d; \
    queueTraceBuffer[queueTraceIndex].pstate = info->si->port_state[info->portnum]; \
    ++queueTraceIndex; \
    queueTraceIndex &= (NumQueueTrace-1); \
  } while (0)

static char *queueEventString[] =
{
  "none",
  "SYNC",
  "SEND",
  "FLSH",
  "QIN",
};

// Implements the /proc/driver/nslink/queue-trace method.
static int queue_proc_read_func(char *buf, char **start, off_t offset, 
                          int count, int *eof, void *data)
{
  struct queueTraceStruct *p;
  int i;
  static unsigned readIndex;
  int cnt = 0;
  int len = 0;
  
  *eof = 0;

  if (offset == 0)
    readIndex = (queueTraceIndex+1)&NumQueueMask;
  
  while (len < (count-128))
    {
      signed short s;
      i = (readIndex+offset+cnt)&NumQueueMask;
      if (i == queueTraceIndex)
        {
          *eof = 1;
          break;
        }
      p = queueTraceBuffer+i;
      s = p->pstate.nPutRemote - p->pstate.nGetRemote;
      len += sprintf(buf+len,"%d: %lu %-6s %4x  %04x %04x %d\n",
                     p->port,p->timestamp,
                     queueEventString[p->event],
                     p->data,
                     p->pstate.nPutRemote,
                     p->pstate.nGetRemote,
                     s);
      ++cnt;
    }
  
  *start = (char*)cnt;
  return len;
}
#else
#define TraceQueue(info,ev,d) /* nooop */
#endif


#if NSLINK_DEBUG_TRACE_HDLC
#define NumAckTrace 4096 /* must be power of 2 */
#define NumAckMask (NumAckTrace-1)
struct ackTraceStruct
{
  unsigned long timestamp;
  int boxid;
  unsigned char out_snd_index;
  unsigned char in_ack_index;
  unsigned char next_in_index;
  unsigned char rtqCount;
  unsigned int head;
  unsigned int tail;
  unsigned event;
  unsigned char data[8];
}ackTraceBuffer[NumAckTrace];
unsigned ackTraceIndex;
unsigned ackTraceDisabled;

#define TraceHdlc(si,ev,d) \
  do {\
    struct sk_buff *h,*t; \
    if (ackTraceDisabled)  break; \
    h = skb_peek(&si->rtq); \
    t = skb_peek_tail(&si->rtq); \
    ackTraceBuffer[ackTraceIndex].timestamp = jiffies; \
    ackTraceBuffer[ackTraceIndex].boxid = si->conc_id; \
    ackTraceBuffer[ackTraceIndex].out_snd_index = si->out_snd_index; \
    ackTraceBuffer[ackTraceIndex].in_ack_index = si->in_ack_index; \
    ackTraceBuffer[ackTraceIndex].next_in_index = si->next_in_index; \
    ackTraceBuffer[ackTraceIndex].rtqCount = si->rtqCount; \
    ackTraceBuffer[ackTraceIndex].head = h ? ((struct HDLChdr *)(h->h.raw + 3))->out_snd_index : -1; \
    ackTraceBuffer[ackTraceIndex].tail = t ? ((struct HDLChdr *)(t->h.raw + 3))->out_snd_index : -1; \
    ackTraceBuffer[ackTraceIndex].event = ev; \
    ackTraceBuffer[ackTraceIndex].data[0] = d; \
    ++ackTraceIndex; \
    ackTraceIndex &= (NumAckTrace-1); \
  } while (0)

#define TraceHdlcStop(si,ev,d) \
  do {TraceHdlc(si,ev,d); ackTraceDisabled=1;} while (0)

static char *stopString[] =
{
  "none",
  "STOP-BADSEQ",
  "STOP-RETRANS",
  "STOP-BADACK", 
};

static char *eventString[] =
{
  "none",
  "RXACK",
  "RECV",
  "SEND",
  "TXACK",
  "RXBAD",
  "RESND",
};

// Implements the /proc/driver/nslink/hdlc-trace method.
static int ack_proc_read_func(char *buf, char **start, off_t offset, 
                          int count, int *eof, void *data)
{
  struct ackTraceStruct *p;
  int i;
  static unsigned readIndex;
  int cnt = 0;
  int len = 0;
  
  *eof = 0;

  if (offset == 0)
    readIndex = (ackTraceIndex+1)&NumAckMask;
  
  while (len < (count-128))
    {
      i = (readIndex+offset+cnt)&NumAckMask;
      if (i == ackTraceIndex)
        {
          *eof = 1;
          break;
        }
      p = ackTraceBuffer+i;
      len += sprintf(buf+len,"%d: %lu %-6s %02x os=%02x ia=%02x ni=%02x c=%d ",
                     p->boxid,p->timestamp,
                     p->event < 0 ? stopString[p->event] : eventString[p->event],
                     p->data[0],
                     p->out_snd_index, p->in_ack_index, 
                     p->next_in_index,p->rtqCount);
      if (p->head == -1)
        len += sprintf(buf+len,"h=-- ");
      else
        len += sprintf(buf+len,"h=%02x ",p->head&0xff);
      if (p->tail == -1)
        len += sprintf(buf+len,"t=-- ");
      else
        len += sprintf(buf+len,"t=%02x ",p->tail&0xff);
      len += sprintf(buf+len,"\n");
      ++cnt;
    }
  
  *start = (char*)cnt;
  return len;
}
#else
#define TraceHdlc(si,ev,d) /* nooop */
#define TraceHdlcStop(si,ev,d) /* nooop */
#endif


// Implements the /proc/driver/nslink/port method.
static int port_proc_read_func(char *buf, char **start, off_t offset, 
                          int count, int *eof, void *data)
{
#if (1) // new way
  struct nr_port *info;
  struct si_port_state *pstate;
  int cnt = 0, len = 0, i;
  int allInfo = 0;

  while (len < (count - 512)) {
      i = offset + cnt;
      if (i == 0) {
	  len += sprintf(buf + len,"Port Open mode  Baud  Ctrl  FC    TxBytes    RxBytes  ParErr  FramErr  OverErr");
	  if (allInfo)
	      len += sprintf(buf + len,"  ctrl  mcr  err getl putr getr  cntr");
	  len += sprintf(buf + len," \n");
      }
      if (i >= (int)((sizeof nrp_table)/(sizeof nrp_table[0])))  {
          *eof = 1;
          break;
      }
      info = nrp_table[i];
      if (info)  {
          len += sprintf(buf + len,"%3d %4d   %02x %6d   ",
			 i,
                         info->count,
                         info->rsmode,
                         info->baud_rate);

	  // Ctrl ...
	  if (info->control_settings & SC_DATABITS_7)
	      len += sprintf(buf + len,"7");
	  else
	      len += sprintf(buf + len,"8");

	  if (info->control_settings & SC_PARITY_ODD)
	      len += sprintf(buf + len,"O");
	  else if (info->control_settings & SC_PARITY_EVEN)
	      len += sprintf(buf + len,"E");
	  else
	      len += sprintf(buf + len,"N");

	  if (info->control_settings & SC_STOPBITS_2)
	      len += sprintf(buf + len,"2  ");
	  else
	      len += sprintf(buf + len,"1  ");

	  // FC ...
	  if (info->control_settings & SC_FLOW_CTS_AUTO)
	      len += sprintf(buf + len,"HW ");
	  else if (info->control_settings & SC_FLOW_XON_TX_AUTO)
	      len += sprintf(buf + len,"SW ");
	  else
	      len += sprintf(buf + len,"-- ");

	  //  TX RX Errors
          len += sprintf(buf + len,"  %8d   %8d  %6d   %6d   %6d  ",
			 info->txBytes,
                         info->rxBytes,
                         info->parityErrors,
                         info->framingErrors,
			 info->overflowErrors);

          if (allInfo) {
	      pstate = info->si->port_state+info->line;
              len += sprintf(buf+len,"%04x %4x %4x %4x %5d",
                             pstate->errorStatusRegister,
                             pstate->nGetLocal,
                             pstate->nPutRemote,
                             pstate->nGetRemote,
                             get_tx_cnt_remote(pstate));
	  }
          len += sprintf(buf+len,"\n");
      }
      ++cnt;
  }
  *start = (char*)cnt;
  return len;


#else // old way

  struct nr_port *info;
  struct si_port_state *pstate;
  int cnt=0;
  int len=0;

  while (len < (count-512))
    {
      int i = offset+cnt;
      if (i==0)
        len += sprintf(buf+len,"port open mode  baud ctrl  mcr  err getl putr getr  cntr\n");
      if (i >= (int)((sizeof nrp_table)/(sizeof nrp_table[0])))
        {
          *eof = 1;
          break;
        }
      info = nrp_table[i];
      if (info)
        {
          len += sprintf(buf+len,"%3d %4d   %02x %6d %04x %04x ",i,
                         info->count,
                         info->rsmode,
                         info->baud_rate,
                         info->control_settings,
                         info->modem_control_register);
          if (info->si && (pstate=info->si->port_state+info->line))
            {
              len += sprintf(buf+len,"%04x %4x %4x %4x %5d",
                             pstate->errorStatusRegister,
                             pstate->nGetLocal,
                             pstate->nPutRemote,
                             pstate->nGetRemote,
                             get_tx_cnt_remote(pstate)
                             );
            }
          len += sprintf(buf+len,"\n");
        }
      ++cnt;
    }
  *start = (char*)cnt;
  return len;

#endif  // old way

}


// Implements the /proc/driver/nslink/hdlc method.
static int hdlc_proc_read_func(char *buf, char **start, off_t offset, 
                          int count, int *eof, void *data)
{
  int len = 0;
  struct sk_buff *h,*t;
  int j,cnt=0;
  struct si_state *p;

  while (len < (count-512))
    {
      int i = offset+cnt;
      if (i >= MAX_NRP_SIS)
        {
          *eof = 1;
          break;
        }
      p = conc_state[i];
      if (!p)
        len += sprintf(buf+len,"%d: xx:xx:xx:xx:xx:xx\n",i);
      else
        {
          len += sprintf(buf+len,"%d: %02x:%02x:%02x:%02x:%02x:%02x\n",i,p->ether_addr[0],p->ether_addr[1],p->ether_addr[2],p->ether_addr[3],p->ether_addr[4],p->ether_addr[5]);
          len += sprintf(buf+len," state %08x\n",p->hdlc_state_flags);
          len += sprintf(buf+len," out_snd_index %02x\n",p->out_snd_index);
          len += sprintf(buf+len," in_ack_index %02x\n",p->in_ack_index);
          len += sprintf(buf+len," next_in_index %02x\n",p->next_in_index);
          len += sprintf(buf+len," rtqCount %d\n", p->rtqCount);
          h = skb_peek(&p->rtq);
          t = skb_peek_tail(&p->rtq);
          for (j=0; j<16 && h; ++j)
            {
              len += sprintf(buf+len,"  %02x\n",((struct HDLChdr*)(h->h.raw+3))->out_snd_index);
              if (!t || t==h)
                break;
              h = h->next;
            }
        }
      ++cnt;
    }
  *start = (char*)cnt;
  return len;
}


// Implements the /proc/driver/nslink/status method.
static int proc_read_func(char *buf, char **start, off_t offset, 
                          int count, int *eof, void *data)
{
  int i,n,len;
  struct si_state *si;
  #define LIMIT (PAGE_SIZE-81)
  
  len = 0;

  len += sprintf(buf+len, "Comtrol NS-Link -- version %s (%s)\n",SI_VERSION,SI_DATE);
  len += sprintf(buf+len, "Built %s  %s\n\n",__DATE__,__TIME__);
  len += sprintf(buf+len, "Hub State        Address            Ports\n");
  len += sprintf(buf+len, "--- ------------ ------------------ -----\n");

  n=0;
  
  for (i=0; i<MAX_NRP_SIS; ++i)
    {
      if (len > (int)LIMIT)
        {
          len += sprintf(buf+len,"[output truncated]\n");
          break;
        }
      
      si = conc_state[i];
      
      if (!si)
        continue;
      
      len += sprintf(buf+len, "%3d %-12s",i,siStateStr(si->state));
      len += sprintf(buf+len, " %-18s",si->use_tcp ? ipToStr(si->ip_addr) : macToStr(si->ether_addr));
      len += sprintf(buf+len, " %2d-%-2d",si->base_port,si->base_port+si->num_ports-1);
      
      len += sprintf(buf+len, "\n");
        
      ++n;
    }
  
  len += sprintf(buf+len, "\n%d hub%s running.\n",n,n==1?"":"s");
  *eof = (len <= (int)LIMIT);
  // len += sprintf(buf+len, "[%d]\n",len);
  return len;
}

/********************************************************************************/
/*                                                                              */
/*        MODULE INIT/CLEANUP FUNCTIONS                                         */
/*                                                                              */
/********************************************************************************/

#if LINUX_VERSION_CODE < VERSION_CODE(2,4,0)
// 2.2 proc stuff
extern inline void create_proc_read_entry (const char *name, mode_t mode,
                struct proc_dir_entry *base, void *read_func, void *data)
{
        struct proc_dir_entry *ent = create_proc_entry (name, 0, NULL);
        if (ent)
                ent->read_proc = read_func;
}

#define proc_mkdir(buf, dir) create_proc_entry(buf, S_IFDIR, dir)

#endif

#ifdef MODULE


/*  
 *  For kernels 2.4 and greater, we don't use USE_COUNTS, we register .owner with THIS_MODULE,
 *  then the system keeps track of use counts for use.  We do use this can_unload function to 
 *  avoid having the module unloaded while a serial port is open.
 *  Returns:  0 if we can unload the module, else 1
 */
#if LINUX_VERSION_CODE > VERSION_CODE(2,3,0)
static int can_unload(void)
{
    if (atomic_read(&nrp_num_ports_open)) {
	printk(KERN_INFO "nslink: can_unload() NOT OK\n");
	return 1;
    }
    else {
	printk(KERN_INFO "nslink: can_unload() OK\n");
	return 0;
    }
}
#endif

#if LINUX_VERSION_CODE > VERSION_CODE(2,4,99)
static struct tty_operations nslink_ops = {
        .open = nrp_open,
        .close = nrp_close,
        .write = nrp_write,
        .put_char = nrp_put_char,
        .write_room = nrp_write_room,
        .chars_in_buffer = nrp_chars_in_buffer,
        .flush_buffer = nrp_flush_buffer,
        .ioctl = nrp_ioctl,
        .throttle = nrp_throttle,
        .unthrottle = nrp_unthrottle,
        .set_termios = nrp_set_termios,
        .stop = nrp_stop,
        .start = nrp_start,
        .hangup = nrp_hangup,
        .break_ctl = nrp_break,
        .send_xchar = nrp_send_xchar,
        .wait_until_sent = nrp_wait_until_sent,
        .tiocmget = nrp_tiocmget,
        .tiocmset = nrp_tiocmset,
        .read_proc = nrp_read_proc,
//	.can_unload = can_unload,
};
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
int init_module(void)
#else
int nrp_init(void)
#endif
{
  int retval;
  static struct proc_dir_entry *procdir;
    
  printk(KERN_INFO "nslink: NS-Link driver, version %s %s\n",SI_VERSION, SI_DATE);
  printk(KERN_INFO "nslink: Built %s %s\n", __DATE__, __TIME__);

#if LINUX_VERSION_CODE > VERSION_CODE(2,3,0)
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
  __this_module.can_unload = can_unload;
#endif
#endif

  /*
   * Initialize the array of pointers to our own internal state
   * structures.
   */
  memset(nrp_table, 0, sizeof(nrp_table));
  
  /*
   * Set up the tty driver structure and then register this
   * driver with the tty layer.
   */
//#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
  memset(&nslink_driver, 0, sizeof(struct tty_driver));
  nslink_driver.num = (MAX_NRP_SIS * MAX_NRP_PORTS);
//#endif
 
  nslink_driver.magic = TTY_DRIVER_MAGIC;
  nslink_driver.flags = TTY_DRIVER_NO_DEVFS;
  nslink_driver.driver_name = "NS-Link";
  nslink_driver.name = "ttySI";
  nslink_driver.major = 0; // dynamically allocated major number
  nslink_driver.minor_start = 0;
  nslink_driver.type = TTY_DRIVER_TYPE_SERIAL;
  nslink_driver.subtype = SERIAL_TYPE_NORMAL;
  nslink_driver.init_termios = tty_std_termios;
  nslink_driver.init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
  nslink_driver.flags = TTY_DRIVER_REAL_RAW;
#if LINUX_VERSION_CODE > VERSION_CODE(2,4,99)
  nslink_driver.devfs_name = "tts/SI";
#endif
#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
  nslink_driver.refcount = &nslink_refcount;
  nslink_driver.table = nslink_table;
  nslink_driver.termios = nslink_termios;
  nslink_driver.termios_locked = nslink_termios_locked;

  nslink_driver.open = nrp_open;
  nslink_driver.close = nrp_close;
  nslink_driver.write = nrp_write;
  nslink_driver.put_char = nrp_put_char;
  nslink_driver.flush_chars = nrp_flush_chars;
  nslink_driver.write_room = nrp_write_room;
  nslink_driver.chars_in_buffer = nrp_chars_in_buffer;
  nslink_driver.flush_buffer = nrp_flush_buffer;
  nslink_driver.ioctl = nrp_ioctl;
  nslink_driver.throttle = nrp_throttle;
  nslink_driver.unthrottle = nrp_unthrottle;
  nslink_driver.set_termios = nrp_set_termios;
  nslink_driver.stop = nrp_stop;
  nslink_driver.start = nrp_start;
  nslink_driver.hangup = nrp_hangup;
  nslink_driver.break_ctl = nrp_break;
  nslink_driver.read_proc = nrp_read_proc;
  nslink_driver.send_xchar = nrp_send_xchar;
  nslink_driver.wait_until_sent = nrp_wait_until_sent;
#endif

#if (LINUX_VERSION_CODE > VERSION_CODE(2,5,0))
//  nslink_driver.refcount = nslink_refcount;
//  nslink_driver.ttys = nslink_table;

//  nslink_driver.tiocmget = nrp_tiocmget;
//  nslink_driver.tiocmset = nrp_tiocmset;
  tty_set_operations(&nslink_driver, &nslink_ops);
  nslink_driver.owner = THIS_MODULE;
#endif /* Kernel > 2.5 */

        retval = tty_register_driver(&nslink_driver);
        if (retval < 0) {
                printk(KERN_INFO "Couldn't install tty NS-Link driver (error %d)\n", -retval);
#if LINUX_VERSION_CODE > VERSION_CODE(2,4,99)
                put_tty_driver(&nslink_driver);
#endif
                return -1;
        }
        printk(KERN_INFO "NS-Link driver is major %d\n", nslink_driver.major);


  // Register the control driver NSLinkctl, used by the daemon to talk to the driver
  control_major = register_chrdev(0, ctl_name, &ctl_fops);
  if (control_major < 0)
    {
      printk(KERN_INFO "nslink: Failed to register the control registry.\n");
      return (control_major);
    }
  // Create the /proc/driver/nslink entries. 
  procdir = proc_mkdir("driver/nslink",NULL);
  if (procdir)
    {
#if (LINUX_VERSION_CODE > VERSION_CODE(2,3,0))
      procdir->owner = THIS_MODULE;
#endif  //  Kernel > 2.2
      create_proc_read_entry ("status", 0, procdir, proc_read_func, NULL);
      create_proc_read_entry("hdlc", 0, procdir, hdlc_proc_read_func, NULL);
      create_proc_read_entry("ports", 0, procdir, port_proc_read_func, NULL);
# if NSLINK_DEBUG_TRACE_HDLC  
      create_proc_read_entry("hdlc-trace", 0, procdir, ack_proc_read_func, NULL);
# endif  
# if NSLINK_DEBUG_TRACE_QUEUE
      create_proc_read_entry("queue-trace", 0, procdir, queue_proc_read_func, NULL);
# endif  
    }

  // Add the protocol handler for 0x11FE packets to the networking stack
  dev_add_pack(&comtrol_packet_type);

  memset(&conc_state, 0, sizeof(conc_state));
  return 0;
}

#if LINUX_VERSION_CODE < VERSION_CODE(2,5,0)
void cleanup_module(void)
#else
static void nrp_cleanup_module(void)
#endif
{
  int retval, i;

  remove_proc_entry("driver/nslink/status",0);
  remove_proc_entry("driver/nslink/hdlc",0);
  remove_proc_entry("driver/nslink/ports",0);
  remove_proc_entry("driver/nslink/hdlc-trace",0);
  remove_proc_entry("driver/nslink/queue-trace",0);
  remove_proc_entry("driver/nslink",0);

  dev_remove_pack(&comtrol_packet_type);

  if (control_major > 0)
    {
      retval = unregister_chrdev(control_major, ctl_name);
      if (retval)
	  printk(KERN_INFO "nslink: Error %d while trying to unregister "
		 "control device.\n", -retval);
      else
	control_major = 0;
    }

  retval = tty_unregister_driver(&nslink_driver);
  if (retval)
    printk(KERN_INFO "nslink: Error %d while trying to unregister "
	   "Nslink driver\n", -retval);
  
  for (i = 0; i < MAX_NRP_SIS; i++)
    free_si_box(i);

  return;
}

/*
 *  Returns 1 if the device is an RS232 device only (DeviceMaster Serial Hub),
 *  else 0.
 */
static int UnitHasNo485(int modelID)
{
    switch (modelID) {
    case 5002115:
    case 5002315:
	return 1;
    default:
	return 0;
    }
}

#endif /* MODULE */
